import datetime
import logging
from typing import List, Optional, Tuple, Sequence

import jinja2
import pytest
from dateutil import parser

from metricflow.engine.metricflow_engine import MetricFlowEngine, MetricFlowQueryRequest
from metricflow.model.semantic_model import SemanticModel
from metricflow.object_utils import assert_values_exhausted
from metricflow.plan_conversion.column_resolver import (
    DefaultColumnAssociationResolver,
)
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.sql_exprs import (
    SqlTimeDeltaExpression,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlDateTruncExpression,
    SqlCastToTimestampExpression,
    SqlStringExpression,
)
from metricflow.time.time_granularity import TimeGranularity
from metricflow.test.compare_df import assert_dataframes_equal
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.integration.configured_test_case import (
    IntegrationTestModel,
    RequiredDwEngineFeatures,
    CONFIGURED_INTEGRATION_TESTS_REPOSITORY,
)
from metricflow.test.test_utils import as_datetime
from metricflow.test.time.configurable_time_source import (
    ConfigurableTimeSource,
)

logger = logging.getLogger(__name__)


class CheckQueryHelpers:
    """Functions that can be used to help render check queries in integration tests."""

    def __init__(self, sql_client: SqlClient) -> None:  # noqa:D
        self._sql_client = sql_client

    def render_time_constraint(
        self,
        expr: str,
        start_time: str,
        stop_time: str,
    ) -> str:
        """Render an expression like "ds >='2020-01-01' AND ds < '2020-01-02'" for start_time = stop_time = '2020-01-01'."""

        start_expr = self.cast_to_ts(f"{start_time}")
        time_format = "%Y-%m-%d"
        stop_time_plus_one_day = (
            datetime.datetime.strptime(stop_time, time_format) + datetime.timedelta(days=1)
        ).strftime(time_format)
        stop_expr = self.cast_to_ts(f"{stop_time_plus_one_day}")
        return f"{expr} >= {start_expr} AND {expr} < {stop_expr}"

    def cast_expr_to_ts(self, expr: str) -> str:
        """Returns the expression as a new expression cast to the timestamp type, if applicable for the DB."""
        if self._sql_client.sql_engine_attributes.timestamp_type_supported:
            return f"CAST({expr} AS {self._sql_client.sql_engine_attributes.timestamp_type_name})"

        return expr

    def cast_to_ts(self, string_literal: str) -> str:
        """Similar to cast_expr_to_ts, but assumes the input string is to be converted into a string literal."""
        return self.cast_expr_to_ts(f"'{string_literal}'")

    def render_date_sub(
        self,
        table_alias: str,
        column_alias: str,
        count: int,
        granularity: TimeGranularity,
    ) -> str:
        """Renders a date subtract expression"""
        expr = SqlTimeDeltaExpression(
            arg=SqlColumnReferenceExpression(SqlColumnReference(table_alias, column_alias)),
            count=count,
            granularity=granularity,
        )
        return self._sql_client.sql_engine_attributes.sql_query_plan_renderer.expr_renderer.render_sql_expr(expr).sql

    def render_date_trunc(self, expr: str, granularity: TimeGranularity) -> str:
        """Return the DATE_TRUNC() call that can be used for converting the given expr to the granularity."""
        renderable_expr = SqlDateTruncExpression(
            time_granularity=granularity,
            arg=SqlCastToTimestampExpression(
                arg=SqlStringExpression(
                    sql_expr=expr,
                    requires_parenthesis=False,
                )
            ),
        )
        return self._sql_client.sql_engine_attributes.sql_query_plan_renderer.expr_renderer.render_sql_expr(
            renderable_expr
        ).sql

    @property
    def double_data_type_name(self) -> str:
        """Return the name of the double data type for the relevant SQL engine"""
        return self._sql_client.sql_engine_attributes.double_data_type_name


def filter_not_supported_features(
    sql_client: SqlClient, required_features: Tuple[RequiredDwEngineFeatures, ...]
) -> Sequence[RequiredDwEngineFeatures]:
    """Given a list of required features, return a list of features not supported by the given SQLClient."""
    not_supported_features: List[RequiredDwEngineFeatures] = []
    for required_feature in required_features:
        if required_feature is RequiredDwEngineFeatures.DATE_TRUNC:
            if not sql_client.sql_engine_attributes.date_trunc_supported:
                not_supported_features.append(required_feature)
        elif required_feature is RequiredDwEngineFeatures.FULL_OUTER_JOIN:
            if not sql_client.sql_engine_attributes.full_outer_joins_supported:
                not_supported_features.append(required_feature)
        else:
            assert_values_exhausted(required_feature)
    return not_supported_features


@pytest.mark.parametrize(
    "name",
    CONFIGURED_INTEGRATION_TESTS_REPOSITORY.all_test_case_names,
    ids=lambda name: f"name={name}",
)
def test_case(
    name: str,
    mf_test_session_state: MetricFlowTestSessionState,
    simple_semantic_model: SemanticModel,
    simple_semantic_model_non_ds: SemanticModel,
    composite_identifier_semantic_model: SemanticModel,
    unpartitioned_multi_hop_join_semantic_model: SemanticModel,
    multi_hop_join_semantic_model: SemanticModel,
    extended_date_semantic_model: SemanticModel,
    sql_client: SqlClient,
    create_simple_model_tables: bool,
    create_message_source_tables: bool,
    create_bridge_table: bool,
    create_extended_date_model_tables: bool,
    time_spine_source: TimeSpineSource,
) -> None:
    """Runs all integration tests configured in the test case YAML directory."""
    case = CONFIGURED_INTEGRATION_TESTS_REPOSITORY.get_test_case(name)
    logger.info(f"Running integration test case: '{case.name}' from file '{case.file_path}'")

    missing_required_features = filter_not_supported_features(sql_client, case.required_features)
    if missing_required_features:
        pytest.skip(f"DW does not support {missing_required_features}")

    semantic_model: Optional[SemanticModel] = None
    if case.model is IntegrationTestModel.SIMPLE_MODEL:
        semantic_model = simple_semantic_model
    elif case.model is IntegrationTestModel.SIMPLE_MODEL_NON_DS:
        semantic_model = simple_semantic_model_non_ds
    elif case.model is IntegrationTestModel.COMPOSITE_IDENTIFIER_MODEL:
        semantic_model = composite_identifier_semantic_model
    elif case.model is IntegrationTestModel.UNPARTITIONED_MULTI_HOP_JOIN_MODEL:
        semantic_model = unpartitioned_multi_hop_join_semantic_model
    elif case.model is IntegrationTestModel.PARTITIONED_MULTI_HOP_JOIN_MODEL:
        semantic_model = multi_hop_join_semantic_model
    elif case.model is IntegrationTestModel.EXTENDED_DATE_MODEL:
        semantic_model = extended_date_semantic_model
    else:
        assert_values_exhausted(case.model)

    assert semantic_model

    engine = MetricFlowEngine(
        semantic_model=semantic_model,
        sql_client=sql_client,
        column_association_resolver=DefaultColumnAssociationResolver(semantic_model),
        time_source=ConfigurableTimeSource(as_datetime("2021-01-04")),
        time_spine_source=time_spine_source,
        system_schema=mf_test_session_state.mf_system_schema,
    )

    check_query_helpers = CheckQueryHelpers(sql_client)

    query_result = engine.query(
        MetricFlowQueryRequest.create_with_random_request_id(
            metric_names=case.metrics,
            group_by_names=case.group_bys,
            limit=case.limit,
            time_constraint_start=parser.parse(case.time_constraint[0]) if case.time_constraint else None,
            time_constraint_end=parser.parse(case.time_constraint[1]) if case.time_constraint else None,
            where_constraint=jinja2.Template(case.where_constraint, undefined=jinja2.StrictUndefined,).render(
                source_schema=mf_test_session_state.mf_source_schema,
                render_time_constraint=check_query_helpers.render_time_constraint,
                TimeGranularity=TimeGranularity,
                render_date_sub=check_query_helpers.render_date_sub,
                render_date_trunc=check_query_helpers.render_date_trunc,
                mf_time_spine_source=time_spine_source.spine_table.sql,
                double_data_type_name=check_query_helpers.double_data_type_name,
            )
            if case.where_constraint
            else None,
            order_by_names=case.order_bys,
        )
    )

    actual = query_result.result_df

    expected = sql_client.query(
        jinja2.Template(case.check_query, undefined=jinja2.StrictUndefined,).render(
            source_schema=mf_test_session_state.mf_source_schema,
            render_time_constraint=check_query_helpers.render_time_constraint,
            TimeGranularity=TimeGranularity,
            render_date_sub=check_query_helpers.render_date_sub,
            render_date_trunc=check_query_helpers.render_date_trunc,
            mf_time_spine_source=time_spine_source.spine_table.sql,
            double_data_type_name=check_query_helpers.double_data_type_name,
        )
    )
    # If we sort, it's effectively not checking the order whatever order that the output was would be overwritten.
    assert_dataframes_equal(actual, expected, sort_columns=not case.check_order, allow_empty=case.allow_empty)



from typing import List
from metricflow.dataflow.sql_table import SqlTable
from metricflow.instances import DataSourceElementReference


from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    ModelValidationRule,
    ValidationError,
    ValidationIssueType,
    validate_safely,
)

# A non-exaustive tuple of reserved keywords
# This list was created by running an intersection of keywords for redshift,
# postgres, bigquery, and snowflake
RESERVED_KEYWORDS = (
    "AND",
    "AS",
    "CREATE",
    "DISTINCT",
    "FOR",
    "FROM",
    "FULL",
    "HAVING",
    "IN",
    "INNER",
    "INTO",
    "IS",
    "JOIN",
    "LEFT",
    "LIKE",
    "NATURAL",
    "NOT",
    "NULL",
    "ON",
    "OR",
    "RIGHT",
    "SELECT",
    "UNION",
    "USING",
    "WHERE",
    "WITH",
)


class ReservedKeywordsRule(ModelValidationRule):
    """Check that any element that ends up being selected by name (instead of expr) isn't a commonly reserved keyword.

    Note: This rule DOES NOT catch all keywords. That is because keywords are
    engine specific, and semantic validations are not engine specific. I.e. if
    you change your underlying data warehouse engine, semantic validations
    should still pass, but your data warehouse validations might fail. However,
    data warehouse validations are slow in comparison to semantic validation
    rules. Thus this rule is intended to catch words that are reserved keywords
    in all supported engines and to fail fast. E.g., `USER` is a reserved keyword
    in Redshift but not in all other supported engines. Therefore if one is
    using Redshift and sets a dimension name to `user`, the config would pass
    this rule, but would then fail Data Warehouse Validations.
    """

    @staticmethod
    @validate_safely(whats_being_done="checking that data source sub element names aren't reserved sql keywords")
    def _validate_data_source_sub_elements(data_source: DataSource) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        for dimension in data_source.dimensions:
            if dimension.name.upper() in RESERVED_KEYWORDS:
                issues.append(
                    ValidationError(
                        context=DataSourceElementContext(
                            file_context=FileContext.from_metadata(data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=dimension.name
                            ),
                            element_type=DataSourceElementType.DIMENSION,
                        ),
                        message=f"'{dimension.name}' is an SQL reserved keyword, and thus cannot be used as a dimension 'name'.",
                    )
                )

        for identifier in data_source.identifiers:
            if identifier.is_composite:
                msg = "'{name}' is an SQL reserved keyword, and thus cannot be used as a sub-identifier 'name'"
                names = [sub_ident.name for sub_ident in identifier.identifiers if sub_ident.name is not None]
            else:
                msg = "'{name}' is an SQL reserved keyword, and thus cannot be used as an identifier 'name'"
                names = [identifier.name]

            for name in names:
                if name.upper() in RESERVED_KEYWORDS:
                    issues.append(
                        ValidationError(
                            context=DataSourceElementContext(
                                file_context=FileContext.from_metadata(data_source.metadata),
                                data_source_element=DataSourceElementReference(
                                    data_source_name=data_source.name, element_name=identifier.name
                                ),
                                element_type=DataSourceElementType.IDENTIFIER,
                            ),
                            message=msg.format(name=name),
                        )
                    )

        for measure in data_source.measures:
            if measure.name.upper() in RESERVED_KEYWORDS:
                issues.append(
                    ValidationError(
                        context=DataSourceElementContext(
                            file_context=FileContext.from_metadata(data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=measure.name
                            ),
                            element_type=DataSourceElementType.MEASURE,
                        ),
                        message=f"'{measure.name}' is an SQL reserved keyword, and thus cannot be used as an measure 'name'.",
                    )
                )

        return issues

    @classmethod
    @validate_safely(whats_being_done="checking that data_source sql_tables are not sql reserved keywords")
    def _validate_data_sources(cls, model: UserConfiguredModel) -> List[ValidationIssueType]:
        """Checks names of objects that are not nested."""
        issues: List[ValidationIssueType] = []
        set_keywords = set(RESERVED_KEYWORDS)

        for data_source in model.data_sources:
            if data_source.sql_table is not None:
                set_sql_table_path_parts = set(
                    [part.upper() for part in SqlTable.from_string(data_source.sql_table).parts_tuple]
                )
                keyword_intersection = set_keywords.intersection(set_sql_table_path_parts)

                if len(keyword_intersection) > 0:
                    issues.append(
                        ValidationError(
                            context=DataSourceContext(
                                file_context=FileContext.from_metadata(data_source.metadata),
                                data_source=data_source.reference,
                            ),
                            message=f"'{data_source.sql_table}' contains the SQL reserved keyword(s) {keyword_intersection}, and thus cannot be used for 'sql_table'.",
                        )
                    )
            issues += cls._validate_data_source_sub_elements(data_source=data_source)

        return issues

    @classmethod
    @validate_safely(
        whats_being_done="running model validation ensuring elements that aren't selected via a defined expr don't contain reserved keywords"
    )
    def validate_model(cls, model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        return cls._validate_data_sources(model=model)



from typing import Sequence, List

from metricflow.column_assoc import ColumnAssociation
from metricflow.dataset.dataset import DataSet
from metricflow.instances import (
    InstanceSet,
    DataSourceReference,
    DataSourceElementInstance,
    InstanceSetTransform,
)
from metricflow.object_utils import flatten_nested_sequence
from metricflow.specs import DimensionSpec, TimeDimensionSpec, IdentifierSpec
from metricflow.sql.sql_plan import (
    SqlSelectStatementNode,
)


class SqlDataSet(DataSet):
    """A metric data set along with the associated SQL query node that can be rendered to get those values."""

    def __init__(self, instance_set: InstanceSet, sql_select_node: SqlSelectStatementNode) -> None:
        """Constructor.

        Args:
            instance_set: Describes the instances in the SQL.
            sql_select_node: The SQL that can be rendered to realize the instance set
        """
        self._sql_select_node = sql_select_node
        super().__init__(instance_set=instance_set)

    @property
    def sql_select_node(self) -> SqlSelectStatementNode:
        """Return a SELECT node that can be used to read data from the given SQL table or SQL query"""
        return self._sql_select_node

    def column_associations_for_identifier(
        self,
        identifier_spec: IdentifierSpec,
    ) -> Sequence[ColumnAssociation]:
        """Given the name of the identifier, return the set of columns associated with it in the data set."""
        matching_instances = 0
        column_associations_to_return = None
        for linkable_instance in self.instance_set.identifier_instances:
            if (
                identifier_spec.element_name == linkable_instance.spec.element_name
                and identifier_spec.identifier_links == linkable_instance.spec.identifier_links
            ):
                column_associations_to_return = linkable_instance.associated_columns
                matching_instances += 1

        if matching_instances > 1:
            raise RuntimeError(
                f"More than one instance with spec {identifier_spec} in " f"instance set: {self.instance_set}"
            )

        if not column_associations_to_return:
            raise RuntimeError(f"No instances with spec {identifier_spec} in instance set: {self.instance_set}")

        return column_associations_to_return

    def column_association_for_dimension(
        self,
        dimension_spec: DimensionSpec,
    ) -> ColumnAssociation:
        """Given the name of the dimension, return the set of columns associated with it in the data set."""
        matching_instances = 0
        column_associations_to_return = None
        for dimension_instance in self.instance_set.dimension_instances:
            if dimension_instance.spec == dimension_spec:
                column_associations_to_return = dimension_instance.associated_columns
                matching_instances += 1

        if matching_instances > 1:
            raise RuntimeError(
                f"More than one dimension instance with spec {dimension_spec} in " f"instance set: {self.instance_set}"
            )

        if not column_associations_to_return:
            raise RuntimeError(
                f"No dimension instances with spec {dimension_spec} in instance set: {self.instance_set}"
            )

        return column_associations_to_return[0]

    def column_association_for_time_dimension(
        self,
        time_dimension_spec: TimeDimensionSpec,
    ) -> ColumnAssociation:
        """Given the name of the time dimension, return the set of columns associated with it in the data set."""
        matching_instances = 0
        column_associations_to_return = None
        for time_dimension_instance in self.instance_set.time_dimension_instances:
            if time_dimension_instance.spec == time_dimension_spec:
                column_associations_to_return = time_dimension_instance.associated_columns
                matching_instances += 1

        if matching_instances > 1:
            raise RuntimeError(
                f"More than one time dimension instance with spec {time_dimension_spec} in "
                f"instance set: {self.instance_set}"
            )

        if not column_associations_to_return:
            raise RuntimeError(
                f"No time dimension instances with spec {time_dimension_spec} in instance set: {self.instance_set}"
            )

        return column_associations_to_return[0]

    @property
    def groupable_column_associations(self) -> Sequence[ColumnAssociation]:
        """Return a flattened iterable with all groupable column associations for the current data set"""
        instances = (
            self.instance_set.identifier_instances
            + self.instance_set.dimension_instances
            + self.instance_set.time_dimension_instances
        )
        return flatten_nested_sequence([instance.associated_columns for instance in instances])


class SameDataSourceReferenceChecker(InstanceSetTransform[bool]):
    """Checks to see that all elements in the instance set come from the same data source."""

    def __init__(self, data_source_reference: DataSourceReference) -> None:  # noqa: D
        self._data_source_reference = data_source_reference

    def transform(self, instance_set: InstanceSet) -> bool:  # noqa: D
        combined: List[DataSourceElementInstance] = []
        combined.extend(instance_set.measure_instances)
        combined.extend(instance_set.dimension_instances)
        combined.extend(instance_set.time_dimension_instances)
        combined.extend(instance_set.identifier_instances)

        return all([all([y.is_from(self._data_source_reference) for y in x.defined_from]) for x in combined])



import textwrap

from metricflow.model.objects.common import YamlConfigFile
from metricflow.model.objects.data_source import DataSourceOrigin, MutabilityType
from metricflow.model.objects.elements.identifier import IdentifierType
from metricflow.model.objects.elements.measure import AggregationType
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.parsing.dir_to_model import parse_yaml_files_to_model
from metricflow.time.time_granularity import TimeGranularity


def test_base_data_source_attribute_parsing() -> None:
    """Test for parsing a data source specification without regard for measures, identifiers, or dimensions"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: base_property_test
          mutability:
            type: append_only
            type_params:
              min: minimum_value
              max: maximum_value
              update_cron: "* * 1 * *"
              along: dimension_column
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert data_source.name == "base_property_test"
    assert data_source.origin == DataSourceOrigin.SOURCE  # auto-filled from default, not user-configurable
    assert data_source.mutability.type == MutabilityType.APPEND_ONLY
    assert data_source.mutability.type_params is not None
    assert data_source.mutability.type_params.min == "minimum_value"
    assert data_source.mutability.type_params.max == "maximum_value"
    assert data_source.mutability.type_params.update_cron == "* * 1 * *"
    assert data_source.mutability.type_params.along == "dimension_column"


def test_data_source_metadata_parsing() -> None:
    """Test for asserting that internal metadata is parsed into the DataSource object"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: metadata_test
          mutability:
            type: immutable
        """
    )
    file = YamlConfigFile(filepath="test_dir/inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert data_source.metadata is not None
    assert data_source.metadata.repo_file_path == "test_dir/inline_for_test"
    assert data_source.metadata.file_slice.filename == "inline_for_test"
    expected_metadata_content = textwrap.dedent(
        """\
        name: metadata_test
        mutability:
          type: immutable
        """
    )
    assert data_source.metadata.file_slice.content == expected_metadata_content


def test_data_source_sql_table_parsing() -> None:
    """Test for parsing a data source specification with a sql_table provided"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: sql_table_test
          mutability:
            type: immutable
          sql_table: "some_schema.source_table"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert data_source.sql_table == "some_schema.source_table"


def test_data_source_sql_query_parsing() -> None:
    """Test for parsing a data source specification with a sql_query provided"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: sql_query_test
          mutability:
            type: immutable
          sql_query: "SELECT * FROM some_schema.source_table"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert data_source.sql_query == "SELECT * FROM some_schema.source_table"


def test_data_source_dbt_model_parsing() -> None:
    """Test for parsing a data source specification with a dbt model provided"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: dbt_model_test
          mutability:
            type: immutable
          dbt_model: "dbt_source.some_model"
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert data_source.dbt_model == "dbt_source.some_model"


def test_data_source_identifier_parsing() -> None:
    """Test for parsing a basic identifier out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: identifier_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          identifiers:
            - name: example_identifier
              type: primary
              role: test_role
              entity: other_identifier
              expr: example_id
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.identifiers) == 1
    identifier = data_source.identifiers[0]
    assert identifier.name == "example_identifier"
    assert identifier.type is IdentifierType.PRIMARY
    assert identifier.role == "test_role"
    assert identifier.entity == "other_identifier"
    assert identifier.expr == "example_id"


def test_data_source_identifier_default_entity_parsing() -> None:
    """Test for parsing an identifier with no entity specified out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: entity_default_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          identifiers:
            - name: example_default_entity_identifier
              type: primary
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.identifiers) == 1
    identifier = data_source.identifiers[0]
    assert identifier.entity == "example_default_entity_identifier"


def test_data_source_composite_sub_identifier_ref_parsing() -> None:
    """Test for parsing a ref-based composite sub-identifier out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: composite_sub_identifier_ref_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          identifiers:
            - name: example_identifier
              type: primary
              identifiers:
                - name: composite_ref_identifier
                  ref: other_identifier
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.identifiers) == 1
    identifier = data_source.identifiers[0]
    assert len(identifier.identifiers) == 1

    ref_sub_identifier = identifier.identifiers[0]
    assert ref_sub_identifier.name == "composite_ref_identifier"
    assert ref_sub_identifier.ref == "other_identifier"
    assert ref_sub_identifier.expr is None


def test_data_source_composite_sub_identifier_expr_parsing() -> None:
    """Test for parsing an expr-based composite sub-identifier out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: composite_sub_identifier_expr_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          identifiers:
            - name: example_identifier
              type: primary
              identifiers:
                - name: composite_expr_identifier
                  expr: CAST(expr_identifier AS BIGINT)
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.identifiers) == 1
    identifier = data_source.identifiers[0]
    assert len(identifier.identifiers) == 1

    expr_sub_identifier = identifier.identifiers[0]
    assert expr_sub_identifier.name == "composite_expr_identifier"
    assert expr_sub_identifier.expr == "CAST(expr_identifier AS BIGINT)"
    assert expr_sub_identifier.ref is None


def test_data_source_measure_parsing() -> None:
    """Test for parsing a measure out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: measure_parsing_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          measures:
            - name: example_measure
              agg: count_distinct
              expr: example_input
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.measures) == 1
    measure = data_source.measures[0]
    assert measure.name == "example_measure"
    assert measure.agg is AggregationType.COUNT_DISTINCT
    assert measure.create_metric is not True
    assert measure.expr == "example_input"


def test_data_source_measure_metadata_parsing() -> None:
    """Test for parsing metadata for a measure object defined in a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: measure_metadata_parsing_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          measures:
            - name: example_measure_with_metadata
              agg: count_distinct
              expr: example_input
        """
    )
    file = YamlConfigFile(filepath="test_dir/inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.measures) == 1
    measure = data_source.measures[0]
    assert measure.metadata is not None
    assert measure.metadata.repo_file_path == "test_dir/inline_for_test"
    assert measure.metadata.file_slice.filename == "inline_for_test"
    expected_metadata_content = textwrap.dedent(
        """\
      name: example_measure_with_metadata
      agg: count_distinct
      expr: example_input
      """
    )
    assert measure.metadata.file_slice.content == expected_metadata_content


def test_data_source_create_metric_measure_parsing() -> None:
    """Test for parsing a measure out of a data source specification when create metric is set"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: measure_parsing_create_metric_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          measures:
            - name: example_measure
              agg: count_distinct
              create_metric: true
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.measures) == 1
    measure = data_source.measures[0]
    assert measure.create_metric is True


def test_data_source_categorical_dimension_parsing() -> None:
    """Test for parsing a categorical dimension out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: dimension_parsing_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          dimensions:
            - name: example_categorical_dimension
              type: categorical
              expr: dimension_input
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.dimensions) == 1
    dimension = data_source.dimensions[0]
    assert dimension.name == "example_categorical_dimension"
    assert dimension.type is DimensionType.CATEGORICAL
    assert dimension.is_partition is not True


def test_data_source_partition_dimension_parsing() -> None:
    """Test for parsing a partition dimension out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: dimension_parsing_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          dimensions:
            - name: example_categorical_dimension
              type: categorical
              is_partition: true
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.dimensions) == 1
    dimension = data_source.dimensions[0]
    assert dimension.is_partition is True


def test_data_source_time_dimension_parsing() -> None:
    """Test for parsing a time dimension out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: dimension_parsing_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          dimensions:
            - name: example_time_dimension
              type: time
              type_params:
                time_format: "%Y-%M-%D"
                time_granularity: month
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.dimensions) == 1
    dimension = data_source.dimensions[0]
    assert dimension.type is DimensionType.TIME
    assert dimension.type_params is not None
    assert dimension.type_params.is_primary is not True
    assert dimension.type_params.time_format == "%Y-%M-%D"
    assert dimension.type_params.time_granularity is TimeGranularity.MONTH


def test_data_source_primary_time_dimension_parsing() -> None:
    """Test for parsing a primary time dimension out of a data source specification"""
    yaml_contents = textwrap.dedent(
        """\
        data_source:
          name: dimension_parsing_test
          mutability:
            type: immutable
          sql_table: some_schema.source_table
          dimensions:
            - name: example_time_dimension
              type: time
              type_params:
                time_granularity: month
                is_primary: true
        """
    )
    file = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)

    build_result = parse_yaml_files_to_model(files=[file])

    assert len(build_result.model.data_sources) == 1
    data_source = build_result.model.data_sources[0]
    assert len(data_source.dimensions) == 1
    dimension = data_source.dimensions[0]
    assert dimension.type is DimensionType.TIME
    assert dimension.type_params is not None
    assert dimension.type_params.is_primary is True



from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Set, Optional, Sequence, Tuple, FrozenSet

from metricflow.errors.errors import (
    DuplicateMetricError,
    MetricNotFoundError,
    NonExistentMeasureError,
    InvalidDataSourceError,
)
from metricflow.instances import DataSourceReference, DataSourceElementReference
from metricflow.model.objects.data_source import DataSource, DataSourceOrigin
from metricflow.model.objects.elements.dimension import Dimension
from metricflow.model.objects.elements.identifier import Identifier
from metricflow.model.objects.elements.measure import Measure, AggregationType, NonAdditiveDimensionParameters
from metricflow.model.objects.metric import Metric, MetricType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantics.data_source_container import PydanticDataSourceContainer
from metricflow.model.semantics.element_group import ElementGrouper
from metricflow.model.semantics.linkable_spec_resolver import (
    ValidLinkableSpecResolver,
    LinkableElementProperties,
)
from metricflow.references import (
    DimensionReference,
    IdentifierReference,
    LinkableElementReference,
    MeasureReference,
    TimeDimensionReference,
)
from metricflow.specs import (
    LinkableInstanceSpec,
    MeasureSpec,
    MetricSpec,
)

logger = logging.getLogger(__name__)

MAX_JOIN_HOPS = 2


class MetricSemantics:  # noqa: D
    def __init__(  # noqa: D
        self, user_configured_model: UserConfiguredModel, data_source_semantics: DataSourceSemantics
    ) -> None:
        self._user_configured_model = user_configured_model
        self._metrics: Dict[MetricSpec, Metric] = {}
        self._data_source_semantics = data_source_semantics

        # Dict from the name of the metric to the hash.
        self._metric_hashes: Dict[MetricSpec, str] = {}

        for metric in self._user_configured_model.metrics:
            self.add_metric(metric)

        self._linkable_spec_resolver = ValidLinkableSpecResolver(
            user_configured_model=self._user_configured_model,
            max_identifier_links=MAX_JOIN_HOPS,
        )

    def element_specs_for_metrics(
        self,
        metric_specs: List[MetricSpec],
        with_any_property: FrozenSet[LinkableElementProperties] = LinkableElementProperties.all_properties(),
        without_any_property: FrozenSet[LinkableElementProperties] = frozenset(),
    ) -> List[LinkableInstanceSpec]:
        """Dimensions common to all metrics requested (intersection)"""

        all_linkable_specs = self._linkable_spec_resolver.get_linkable_elements_for_metrics(
            metric_specs=metric_specs,
            with_any_of=with_any_property,
            without_any_of=without_any_property,
        ).as_spec_set

        return sorted(all_linkable_specs.as_tuple, key=lambda x: x.qualified_name)

    def get_metrics(self, metric_names: List[MetricSpec]) -> List[Metric]:  # noqa: D
        res = []
        for metric_name in metric_names:
            if metric_name not in self._metrics:
                raise MetricNotFoundError(f"Unable to find metric `{metric_name}`. Perhaps it has not been registered")
            res.append(self._metrics[metric_name])

        return res

    @property
    def metric_names(self) -> List[MetricSpec]:  # noqa: D
        return list(self._metrics.keys())

    def get_metric(self, metric_name: MetricSpec) -> Metric:  # noqa:D
        if metric_name not in self._metrics:
            raise MetricNotFoundError(f"Unable to find metric `{metric_name}`. Perhaps it has not been registered")
        return self._metrics[metric_name]

    def add_metric(self, metric: Metric) -> None:
        """Add metric, validating presence of required measures"""
        metric_spec = MetricSpec(element_name=metric.name)
        if metric_spec in self._metrics:
            raise DuplicateMetricError(f"Metric `{metric.name}` has already been registered")
        for measure_reference in metric.measure_references:
            if measure_reference not in self._data_source_semantics.measure_references:
                raise NonExistentMeasureError(
                    f"Metric `{metric.name}` references measure `{measure_reference}` which has not been registered"
                )
        self._metrics[metric_spec] = metric
        self._metric_hashes[metric_spec] = metric.definition_hash

    @property
    def valid_hashes(self) -> Set[str]:
        """Return all of the hashes of the metric definitions."""
        return set(self._metric_hashes.values())

    def measures_for_metric(self, metric_spec: MetricSpec) -> Tuple[MeasureSpec, ...]:
        """Return the measure specs required to compute the metric."""
        metric = self.get_metric(metric_spec)

        return tuple(
            MeasureSpec(
                element_name=x.element_name,
                non_additive_dimension=self._data_source_semantics.non_additive_dimension_by_measure.get(x),
            )
            for x in metric.measure_references
        )

    def contains_cumulative_metric(self, metric_specs: Sequence[MetricSpec]) -> bool:
        """Returns true if any of the specs correspond to a cumulative metric."""
        for metric_spec in metric_specs:
            if self.get_metric(metric_spec).type == MetricType.CUMULATIVE:
                return True
        return False


class DataSourceSemantics:
    """Tracks semantic information for data source held in a set of DataSourceContainers"""

    def __init__(  # noqa: D
        self,
        model: UserConfiguredModel,
        configured_data_source_container: PydanticDataSourceContainer,
    ) -> None:
        self._model = model
        self._measure_index: Dict[MeasureReference, List[DataSource]] = defaultdict(list)
        self._measure_aggs: Dict[
            MeasureReference, AggregationType
        ] = {}  # maps measures to their one consistent aggregation
        self._measure_non_additive_dimension: Dict[MeasureReference, NonAdditiveDimensionParameters] = {}
        self._dimension_index: Dict[DimensionReference, List[DataSource]] = defaultdict(list)
        self._linkable_reference_index: Dict[LinkableElementReference, List[DataSource]] = defaultdict(list)
        self._entity_index: Dict[Optional[str], List[DataSource]] = defaultdict(list)
        self._identifier_ref_to_entity: Dict[IdentifierReference, Optional[str]] = {}
        self._data_source_names: Set[str] = set()

        self._configured_data_source_container = configured_data_source_container
        self._data_source_to_aggregation_time_dimensions: Dict[
            DataSourceReference, ElementGrouper[TimeDimensionReference, MeasureSpec]
        ] = {}

        # Add semantic tracking for data sources from configured_data_source_container
        for data_source in self._configured_data_source_container.values():
            assert isinstance(data_source, DataSource)
            self.add_configured_data_source(data_source)

    def add_configured_data_source(self, data_source: DataSource) -> None:
        """Dont use this unless you mean it (ie in tests). The configured data sources are supposed to be static"""
        self._configured_data_source_container._put(data_source)
        self._add_data_source(data_source)

    def get_dimension_references(self) -> List[DimensionReference]:  # noqa: D
        return list(self._dimension_index.keys())

    def get_dimension(
        self, dimension_reference: DimensionReference, origin: Optional[DataSourceOrigin] = None
    ) -> Dimension:
        """Retrieves a full dimension object by name"""
        for dimension_source in self._dimension_index[dimension_reference]:
            if origin and dimension_source.origin != origin:
                continue
            dimension = dimension_source.get_dimension(dimension_reference)
            # find the data source that has the requested dimension by the requested identifier

            return deepcopy(dimension)

        raise ValueError(
            f"Could not find dimension with name ({dimension_reference.element_name}) in configured data sources"
        )

    def get_time_dimension(self, time_dimension_reference: TimeDimensionReference) -> Dimension:
        """Retrieves a full dimension object by name"""
        dimension_reference = time_dimension_reference.dimension_reference()

        if dimension_reference not in self._dimension_index:
            raise ValueError(
                f"Could not find dimension with name ({dimension_reference.element_name}) in configured data sources"
            )

        for dimension_source in self._dimension_index[dimension_reference]:
            dimension = dimension_source.get_dimension(dimension_reference)
            # TODO: Unclear if the deepcopy is necessary.
            return deepcopy(dimension)

        assert False, f"{time_dimension_reference} should have been in the dimension index"

    @property
    def measure_references(self) -> List[MeasureReference]:  # noqa: D
        return list(self._measure_index.keys())

    @property
    def non_additive_dimension_by_measure(self) -> Dict[MeasureReference, NonAdditiveDimensionParameters]:  # noqa: D
        return self._measure_non_additive_dimension

    def get_measure(self, measure_reference: MeasureReference) -> Measure:  # noqa: D
        if measure_reference not in self._measure_index:
            raise ValueError(f"Could not find measure with name ({measure_reference}) in configured data sources")

        assert len(self._measure_index[measure_reference]) >= 1
        # Measures should be consistent across data sources, so just use the first one.
        return list(self._measure_index[measure_reference])[0].get_measure(measure_reference)

    def get_identifier_references(self) -> List[IdentifierReference]:  # noqa: D
        return list(self._identifier_ref_to_entity.keys())

    # DSC interface
    def get_data_sources_for_measure(self, measure_reference: MeasureReference) -> List[DataSource]:  # noqa: D
        return self._measure_index[measure_reference]

    def get_identifier_in_data_source(self, ref: DataSourceElementReference) -> Optional[Identifier]:  # Noqa: d
        data_source = self.get(ref.data_source_name)
        if not data_source:
            return None

        for identifier in data_source.identifiers:
            if identifier.reference.element_name == ref.element_name:
                return identifier

        return None

    def get(self, data_source_name: str) -> Optional[DataSource]:  # noqa: D
        if data_source_name in self._configured_data_source_container:
            data_source = self._configured_data_source_container.get(data_source_name)
            assert isinstance(data_source, DataSource)
            return data_source

        return None

    def get_by_reference(self, data_source_reference: DataSourceReference) -> Optional[DataSource]:  # noqa: D
        return self.get(data_source_reference.data_source_name)

    def _add_data_source(
        self,
        data_source: DataSource,
        fail_on_error: bool = True,
        logging_context: str = "",
    ) -> None:
        """Add data source semantic information, validating consistency with existing data sources."""
        errors = []

        if data_source.name in self._data_source_names:
            errors.append(f"name {data_source.name} already registered - please ensure data source names are unique")

        for measure in data_source.measures:
            if measure.reference in self._measure_aggs and self._measure_aggs[measure.reference] != measure.agg:
                errors.append(
                    f"conflicting aggregation (agg) for measure `{measure.reference.element_name}` registered as "
                    f"`{self._measure_aggs[measure.reference]}`; Got `{measure.agg}"
                )

        if errors:
            error_prefix = "\n  - "
            error_msg = (
                f"Unable to add data source `{data_source.name}` "
                f"{'while ' + logging_context + ' ' if logging_context else ''}"
                f"{'... skipping' if not fail_on_error else ''}.\n"
                f"Errors: {error_prefix + error_prefix.join(errors)}"
            )
            if fail_on_error:
                raise InvalidDataSourceError(error_msg)
            logger.warning(error_msg)
            return

        self._data_source_names.add(data_source.name)
        self._data_source_to_aggregation_time_dimensions[data_source.reference] = ElementGrouper[
            TimeDimensionReference, MeasureSpec
        ]()

        for measure in data_source.measures:
            self._measure_aggs[measure.reference] = measure.agg
            self._measure_index[measure.reference].append(data_source)
            agg_time_dimension = measure.checked_agg_time_dimension
            self._data_source_to_aggregation_time_dimensions[data_source.reference].add_value(
                key=agg_time_dimension,
                value=MeasureSpec(
                    element_name=measure.name,
                    non_additive_dimension=measure.non_additive_dimension,
                ),
            )
            if measure.non_additive_dimension:
                self._measure_non_additive_dimension[measure.reference] = measure.non_additive_dimension
        for dim in data_source.dimensions:
            self._linkable_reference_index[dim.reference].append(data_source)
            self._dimension_index[dim.reference].append(data_source)
        for ident in data_source.identifiers:
            self._identifier_ref_to_entity[ident.reference] = ident.entity
            self._entity_index[ident.entity].append(data_source)
            self._linkable_reference_index[ident.reference].append(data_source)

    @property
    def data_source_references(self) -> Sequence[DataSourceReference]:  # noqa: D
        data_source_names_sorted = sorted(self._data_source_names)
        return tuple(DataSourceReference(data_source_name=x) for x in data_source_names_sorted)

    def get_aggregation_time_dimensions_with_measures(
        self, data_source_reference: DataSourceReference
    ) -> ElementGrouper[TimeDimensionReference, MeasureSpec]:
        """Return all time dimensions in a data source with their associated measures."""
        assert (
            data_source_reference in self._data_source_to_aggregation_time_dimensions
        ), f"Data Source {data_source_reference} is not known"
        return self._data_source_to_aggregation_time_dimensions[data_source_reference]



import logging
import os
from dataclasses import dataclass
from string import Template
import traceback
from typing import Optional, Dict, List, Union, Type

from jsonschema import exceptions

from metricflow.errors.errors import ParsingException
from metricflow.model.model_transformer import ModelTransformer
from metricflow.model.objects.common import Version, YamlConfigFile
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.materialization import Materialization
from metricflow.model.objects.metric import Metric
from metricflow.model.parsing.schemas_internal import (
    metric_validator,
    data_source_validator,
    materialization_validator,
)
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.yaml_loader import (
    ParsingContext,
    YamlConfigLoader,
    PARSING_CONTEXT_KEY,
)
from metricflow.model.validations.validator_helpers import (
    FileContext,
    ModelValidationException,
    ModelValidationResults,
    ValidationError,
    ValidationIssueType,
)

logger = logging.getLogger(__name__)

VERSION_KEY = "mf_config_schema"
METRIC_TYPE = "metric"
DATA_SOURCE_TYPE = "data_source"
MATERIALIZATION_TYPE = "materialization"
DOCUMENT_TYPES = [METRIC_TYPE, DATA_SOURCE_TYPE, MATERIALIZATION_TYPE]


@dataclass(frozen=True)
class ModelBuildResult:  # noqa: D
    model: UserConfiguredModel
    # Issues found in the model.
    issues: ModelValidationResults = ModelValidationResults()


@dataclass(frozen=True)
class FileParsingResult:
    """Results of parsing a config file

    Attributes:
        elements: MetricFlow model elements parsed from the file
        issues: Issues found when trying to parse the file
    """

    elements: List[Union[DataSource, Metric, Materialization]]
    issues: List[ValidationIssueType]


def parse_directory_of_yaml_files_to_model(
    directory: str,
    template_mapping: Optional[Dict[str, str]] = None,
    apply_pre_transformations: Optional[bool] = True,
    apply_post_transformations: Optional[bool] = True,
    raise_issues_as_exceptions: bool = True,
) -> ModelBuildResult:
    """Parse files in the given directory to a TMdoModel.

    Strings in the file following the Python string template format are replaced according to the template_mapping dict.
    """
    template_mapping = template_mapping or {}
    yaml_config_files = []

    for root, dirs, files in os.walk(directory):
        # Skip hidden directories. os.walk() supports mutation of dirs to skip directories.
        dirs[:] = [d for d in dirs if not d.startswith(".")]

        for file in files:
            if not YamlConfigLoader.is_valid_yaml_file_ending(file):
                continue
            # Skip hidden files
            if file.startswith("."):
                continue

            file_path = os.path.join(root, file)
            with open(file_path) as f:
                contents = Template(f.read()).substitute(template_mapping)
                yaml_config_files.append(
                    YamlConfigFile(filepath=file_path, contents=contents),
                )

    build_result = parse_yaml_files_to_model(yaml_config_files)
    model = build_result.model
    assert model

    build_issues = build_result.issues
    try:
        if apply_pre_transformations:
            model = ModelTransformer.pre_validation_transform_model(model)

        if apply_post_transformations:
            model = ModelTransformer.post_validation_transform_model(model)
    except Exception as e:
        transformation_issue_results = ModelValidationResults(errors=[ValidationError(message=str(e))])
        build_issues = ModelValidationResults.merge([build_issues, transformation_issue_results])

    if raise_issues_as_exceptions and build_issues.has_blocking_issues:
        raise ModelValidationException(build_issues.all_issues)

    return ModelBuildResult(model=model, issues=build_issues)


def parse_yaml_files_to_model(
    files: List[YamlConfigFile],
    data_source_class: Type[DataSource] = DataSource,
    metric_class: Type[Metric] = Metric,
    materialization_class: Type[Materialization] = Materialization,
) -> ModelBuildResult:
    """Builds UserConfiguredModel from list of config files (as strings).

    Persistent storage connection may be passed to write parsed objects=
    to storage and populate object metadata

    Note: this function does not finalize the model
    """
    data_sources = []
    metrics = []
    materializations: List[Materialization] = []
    valid_object_classes = [data_source_class.__name__, metric_class.__name__, materialization_class.__name__]
    issues: List[ValidationIssueType] = []

    for config_file in files:
        parsing_result = parse_config_yaml(  # parse config file
            config_file,
            data_source_class=data_source_class,
            metric_class=metric_class,
            materialization_class=materialization_class,
        )
        file_issues = parsing_result.issues
        for obj in parsing_result.elements:
            if isinstance(obj, data_source_class):
                data_sources.append(obj)
            elif isinstance(obj, metric_class):
                metrics.append(obj)
            elif isinstance(obj, materialization_class):
                materializations.append(obj)
            else:
                file_issues.append(
                    ValidationError(
                        context=FileContext(file_name=config_file.filepath),
                        message=f"Unexpected model object {obj.__name__}. Expected {valid_object_classes}.",
                    )
                )

        issues += file_issues

    return ModelBuildResult(
        model=UserConfiguredModel(
            data_sources=data_sources,
            materializations=materializations,
            metrics=metrics,
        ),
        issues=ModelValidationResults.from_issues_sequence(issues),
    )


def parse_config_yaml(
    config_yaml: YamlConfigFile,
    data_source_class: Type[DataSource] = DataSource,
    metric_class: Type[Metric] = Metric,
    materialization_class: Type[Materialization] = Materialization,
) -> FileParsingResult:
    """Parses transform config file passed as string - Returns list of model objects"""
    results: List[Union[DataSource, Metric, Materialization]] = []
    ctx: Optional[ParsingContext] = None
    issues: List[ValidationIssueType] = []
    try:
        for config_document in YamlConfigLoader.load_all_with_context(
            name=config_yaml.filepath, contents=config_yaml.contents
        ):
            # The config document can be None if there is nothing but white space between two `---`
            # this isn't really an issue, so lets just swallow it
            if config_document is None:
                continue
            if not isinstance(config_document, dict):
                issues.append(
                    ValidationError(
                        context=FileContext(file_name=config_yaml.filepath),
                        message=f"YAML must be a dict. Got `{type(config_document)}`.",
                    )
                )
                continue

            keys = config_document.keys()

            # This SHOULDN'T ever happen but if it does, we want to know. If this
            # does ever happen, it is likely due to a change in 'load_all_with_context'
            if PARSING_CONTEXT_KEY not in keys:
                raise RuntimeError(
                    f"No parsing context present. Expected key `{PARSING_CONTEXT_KEY}` from the YAML parser."
                )

            ctx = config_document.pop(PARSING_CONTEXT_KEY)
            assert ctx

            if VERSION_KEY in config_document:
                version = Version.parse(config_document.pop(VERSION_KEY))
                major_version = version.major

                if major_version != 0:
                    issues.append(
                        ValidationError(
                            context=FileContext(file_name=ctx.filename, line_number=ctx.start_line),
                            message=f"Unsupported version {version} in config document.",
                        )
                    )

            # Because we've popped the VERSION KEY and PARSING_CONTEXT_KEY, there
            # should only be the base object key remaining
            if len(keys) != 1:
                issues.append(
                    ValidationError(
                        context=FileContext(file_name=ctx.filename, line_number=ctx.start_line),
                        message=f"Document should have one type of key, but has {keys}.",
                    )
                )
                continue

            # retrieve last top-level key as type
            document_type = next(iter(config_document.keys()))
            object_cfg = config_document[document_type]

            try:
                if document_type == METRIC_TYPE:
                    metric_validator.validate(config_document[document_type])
                    results.append(metric_class.parse_obj(object_cfg))
                elif document_type == DATA_SOURCE_TYPE:
                    data_source_validator.validate(config_document[document_type])
                    results.append(data_source_class.parse_obj(object_cfg))
                elif document_type == MATERIALIZATION_TYPE:
                    materialization_validator.validate(config_document[document_type])
                    results.append(materialization_class.parse_obj(object_cfg))
                else:
                    issues.append(
                        ValidationError(
                            context=FileContext(file_name=ctx.filename, line_number=ctx.start_line),
                            message=f"Invalid document type: {document_type}. Expected {DOCUMENT_TYPES}.",
                        )
                    )
            # catches exceptions from jsonschema validator
            except exceptions.ValidationError as e:
                context = FileContext(file_name=ctx.filename, line_number=ctx.start_line)
                issues.append(
                    ValidationError(
                        context=context,
                        message=f"YAML document did not conform to metric spec.\nError: {e}",
                        extra_detail="".join(traceback.format_tb(e.__traceback__)),
                    )
                )
            # ParsingException: catches exceptions from *.parse_obj calls
            # Exception: general exception for a given document. Basicially we
            # don't want an exception on one document to halt checking the rest
            # of the documents
            except (ParsingException, Exception) as e:
                context = FileContext(file_name=ctx.filename, line_number=ctx.start_line)
                issues.append(
                    ValidationError(
                        context=context,
                        message=str(e),
                        extra_detail="".join(traceback.format_tb(e.__traceback__)),
                    )
                )
    # If a runtime error occured, we still want this to break things
    except RuntimeError:
        raise
    # Any other error should be handled as an issue
    except Exception as e:
        context = FileContext(file_name=config_yaml.filepath)
        issues.append(
            ValidationError(context=context, message=str(e), extra_detail="".join(traceback.format_tb(e.__traceback__)))
        )

    return FileParsingResult(elements=results, issues=issues)



import pytest
import click
from click.testing import CliRunner

from metricflow.cli.custom_click_types import SequenceParamType, MutuallyExclusiveOption


def test_check_min_length() -> None:
    """Make sure `SequenceParamType` checks for `min_length` param."""
    ltype = SequenceParamType[str](min_length=2)
    ltype.convert("1,2", None, None)
    with pytest.raises(click.BadParameter):
        ltype.convert("1", None, None)


def test_check_max_length() -> None:
    """Make sure `SequenceParamType` checks for `max_length` param."""
    ltype = SequenceParamType[str](max_length=2)
    ltype.convert("1,2", None, None)
    with pytest.raises(click.BadParameter):
        ltype.convert("1,2,3", None, None)


def test_use_value_converter() -> None:
    """Make sure `SequenceParamType` uses `value_converter`."""
    ltype = SequenceParamType[int](value_converter=lambda x_str: int(x_str))
    values = ltype.convert("1,2,3,4,5", None, None)

    assert values == [1, 2, 3, 4, 5]


def test_value_converter_fail() -> None:
    """Make sure `SequenceParamType` fails gracefully when a value cannot be converted."""
    ltype = SequenceParamType[int](value_converter=lambda x_str: int(x_str))
    with pytest.raises(click.BadParameter):
        ltype.convert("1,abc", None, None)


def test_use_separator() -> None:
    """Make sure `SequenceParamType` uses `separator`."""
    ltype = SequenceParamType[str](separator=":")

    sep_values = ltype.convert("1:2:3:4:5", None, None)
    assert sep_values == ["1", "2", "3", "4", "5"]

    no_sep_values = ltype.convert("1,2,3,4,5", None, None)
    assert no_sep_values == ["1,2,3,4,5"]


def test_mutually_exclusive_option() -> None:
    """Make sure `MutuallyExclusiveOption` works"""

    @click.command()
    @click.option(
        "--foo",
        cls=MutuallyExclusiveOption,
        mutually_exclusive=["bar"],
        help="",
    )
    @click.option(
        "--bar",
        cls=MutuallyExclusiveOption,
        mutually_exclusive=["foo"],
        help="",
    )
    def test(foo: str, bar: str) -> None:
        pass

    runner = CliRunner()

    # should work when exclusivity is respected
    result = runner.invoke(test, ["--foo", "abc"])
    assert result.exit_code == 0
    result = runner.invoke(test, ["--bar", "abc"])
    assert result.exit_code == 0

    # should fail when it is not respected
    result = runner.invoke(test, ["--foo", "abc", "--bar", "abc"])
    assert result.exit_code != 0



from abc import ABC
from typing import List

from metricflow.inference.models import InferenceResult


class InferenceRenderer(ABC):
    """Render inference results into some format."""

    def render(self, results: List[InferenceResult]) -> None:
        """Render a set of inference results into the screen, some file, the network or whatever."""
        raise NotImplementedError



import logging
import textwrap

import pytest

from metricflow.sql.render.expr_renderer import DefaultSqlExpressionRenderer
from metricflow.sql.sql_exprs import (
    SqlStringExpression,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlComparisonExpression,
    SqlComparison,
    SqlFunctionExpression,
    SqlFunction,
    SqlNullExpression,
    SqlLogicalExpression,
    SqlLogicalOperator,
    SqlStringLiteralExpression,
    SqlIsNullExpression,
    SqlDateTruncExpression,
    SqlRatioComputationExpression,
    SqlColumnReplacements,
    SqlCastToTimestampExpression,
    SqlBetweenExpression,
)
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


@pytest.fixture
def default_expr_renderer() -> DefaultSqlExpressionRenderer:  # noqa: D
    return DefaultSqlExpressionRenderer()


def test_str_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(SqlStringExpression("a + b")).sql
    expected = "a + b"
    assert actual == expected


def test_col_ref_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlColumnReferenceExpression(SqlColumnReference("my_table", "my_col"))
    ).sql
    expected = "my_table.my_col"
    assert actual == expected


def test_comparison_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlComparisonExpression(
            left_expr=SqlColumnReferenceExpression(SqlColumnReference("my_table", "my_col")),
            comparison=SqlComparison.EQUALS,
            right_expr=SqlStringExpression("a + b"),
        )
    ).sql
    assert actual == "my_table.my_col = (a + b)"


def test_require_parenthesis(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlComparisonExpression(
            left_expr=SqlColumnReferenceExpression(SqlColumnReference("a", "booking_value")),
            comparison=SqlComparison.GREATER_THAN,
            right_expr=SqlStringExpression("100", requires_parenthesis=False),
        )
    ).sql

    assert actual == "a.booking_value > 100"


def test_function_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlFunctionExpression(
            sql_function=SqlFunction.SUM,
            sql_function_args=[
                SqlColumnReferenceExpression(SqlColumnReference("my_table", "a")),
                SqlColumnReferenceExpression(SqlColumnReference("my_table", "b")),
            ],
        )
    ).sql
    assert actual == "SUM(my_table.a, my_table.b)"


def test_distinct_agg_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:
    """Distinct aggregation functions require the insertion of the DISTINCT keyword in the rendered function expr"""
    actual = default_expr_renderer.render_sql_expr(
        SqlFunctionExpression(
            sql_function=SqlFunction.COUNT_DISTINCT,
            sql_function_args=[
                SqlColumnReferenceExpression(SqlColumnReference("my_table", "a")),
                SqlColumnReferenceExpression(SqlColumnReference("my_table", "b")),
            ],
        )
    ).sql

    assert actual == "COUNT(DISTINCT my_table.a, my_table.b)"


def test_nested_function_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlFunctionExpression(
            sql_function=SqlFunction.CONCAT,
            sql_function_args=[
                SqlColumnReferenceExpression(SqlColumnReference("my_table", "a")),
                SqlFunctionExpression(
                    sql_function=SqlFunction.CONCAT,
                    sql_function_args=[
                        SqlColumnReferenceExpression(SqlColumnReference("my_table", "b")),
                        SqlColumnReferenceExpression(SqlColumnReference("my_table", "c")),
                    ],
                ),
            ],
        )
    ).sql
    assert actual == "CONCAT(my_table.a, CONCAT(my_table.b, my_table.c))"


def test_null_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(SqlNullExpression()).sql
    assert actual == "NULL"


def test_and_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlLogicalExpression(
            operator=SqlLogicalOperator.AND,
            args=(
                SqlStringExpression("1 < 2", requires_parenthesis=True),
                SqlStringExpression("foo", requires_parenthesis=False),
            ),
        )
    ).sql
    assert (
        actual
        == textwrap.dedent(
            """\
            (1 < 2) AND foo
            """
        ).rstrip()
    )


def test_long_and_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlLogicalExpression(
            operator=SqlLogicalOperator.AND,
            args=(
                SqlStringExpression("some_long_expression1"),
                SqlStringExpression("some_long_expression2"),
                SqlStringExpression("some_long_expression3"),
            ),
        )
    ).sql
    assert (
        actual
        == textwrap.dedent(
            """\
            (
              some_long_expression1
            ) AND (
              some_long_expression2
            ) AND (
              some_long_expression3
            )
            """
        ).rstrip()
    )


def test_string_literal_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(SqlStringLiteralExpression("foo")).sql
    assert actual == "'foo'"


def test_is_null_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlIsNullExpression(SqlStringExpression("foo", requires_parenthesis=False))
    ).sql
    assert actual == "foo IS NULL"


def test_date_trunc_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlDateTruncExpression(time_granularity=TimeGranularity.MONTH, arg=SqlStringExpression("ds"))
    ).sql
    assert actual == "DATE_TRUNC('month', ds)"


def test_ratio_computation_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlRatioComputationExpression(
            numerator=SqlFunctionExpression(
                SqlFunction.SUM, sql_function_args=[SqlStringExpression(sql_expr="1", requires_parenthesis=False)]
            ),
            denominator=SqlColumnReferenceExpression(SqlColumnReference(column_name="divide_by_me", table_alias="a")),
        ),
    ).sql
    assert actual == "CAST(SUM(1) AS DOUBLE) / CAST(NULLIF(a.divide_by_me, 0) AS DOUBLE)"


def test_expr_rewrite(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    expr = SqlLogicalExpression(
        operator=SqlLogicalOperator.AND,
        args=(
            SqlColumnReferenceExpression(SqlColumnReference("a", "col0")),
            SqlColumnReferenceExpression(SqlColumnReference("a", "col1")),
        ),
    )

    column_replacements = SqlColumnReplacements(
        {
            SqlColumnReference("a", "col0"): SqlStringExpression("foo", requires_parenthesis=False),
            SqlColumnReference("a", "col1"): SqlStringExpression("bar", requires_parenthesis=False),
        }
    )
    expr_rewritten = expr.rewrite(column_replacements)
    assert default_expr_renderer.render_sql_expr(expr_rewritten).sql == "foo AND bar"


def test_between_expr(default_expr_renderer: DefaultSqlExpressionRenderer) -> None:  # noqa: D
    actual = default_expr_renderer.render_sql_expr(
        SqlBetweenExpression(
            column_arg=SqlColumnReferenceExpression(SqlColumnReference("a", "col0")),
            start_expr=SqlCastToTimestampExpression(
                arg=SqlStringLiteralExpression(
                    literal_value="2020-01-01",
                )
            ),
            end_expr=SqlCastToTimestampExpression(
                arg=SqlStringLiteralExpression(
                    literal_value="2020-01-10",
                )
            ),
        )
    ).sql
    assert actual == "a.col0 BETWEEN CAST('2020-01-01' AS TIMESTAMP) AND CAST('2020-01-10' AS TIMESTAMP)"



import logging

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.materializations import ValidMaterializationRule
from metricflow.test.model.validations.helpers import materialization_with_guaranteed_meta
from metricflow.test.test_utils import (
    model_with_materialization,
)

logger = logging.getLogger(__name__)


def test_materialization_validation(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    ValidMaterializationRule.validate_model(simple_model__pre_transforms)


def test_identifier(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    assert (
        len(
            ValidMaterializationRule.validate_model(
                model_with_materialization(
                    simple_model__pre_transforms,
                    [
                        materialization_with_guaranteed_meta(
                            name="foobar",
                            metrics=["bookings"],
                            dimensions=["metric_time", "listing"],
                        )
                    ],
                )
            )
        )
        == 0
    )


def test_invalid_metric_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    assert (
        len(
            ValidMaterializationRule.validate_model(
                model_with_materialization(
                    simple_model__pre_transforms,
                    [
                        materialization_with_guaranteed_meta(
                            name="foobar",
                            metrics=["invalid_bookings"],
                            dimensions=["metric_time"],
                        )
                    ],
                )
            )
        )
        == 1
    )


def test_invalid_dimension_name(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    assert (
        len(
            ValidMaterializationRule.validate_model(
                model_with_materialization(
                    simple_model__pre_transforms,
                    [
                        materialization_with_guaranteed_meta(
                            name="foobar",
                            metrics=["bookings"],
                            dimensions=["metric_time", "invalid_dimension_name"],
                        )
                    ],
                )
            )
        )
        == 1
    )


def test_missing_primary_time_dimension(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    """Materializations should have the primary time dimension listed as a dimension"""
    assert (
        len(
            ValidMaterializationRule.validate_model(
                model_with_materialization(
                    simple_model__pre_transforms,
                    [
                        materialization_with_guaranteed_meta(
                            name="foobar",
                            metrics=["bookings"],
                            dimensions=["is_instant"],
                        )
                    ],
                )
            )
        )
        == 1
    )


def test_valid_time_granularity(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    assert (
        len(
            ValidMaterializationRule.validate_model(
                model_with_materialization(
                    simple_model__pre_transforms,
                    [
                        materialization_with_guaranteed_meta(
                            name="materialization_test_case",
                            metrics=["bookings"],
                            dimensions=["metric_time__day"],
                        )
                    ],
                )
            )
        )
        == 0
    )


def test_invalid_time_granularity(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    assert (
        len(
            ValidMaterializationRule.validate_model(
                model_with_materialization(
                    simple_model__pre_transforms,
                    [
                        materialization_with_guaranteed_meta(
                            name="materialization_test_case",
                            metrics=["revenue"],
                            dimensions=["metric_time__hour"],
                        )
                    ],
                )
            )
        )
        == 2
    )



from __future__ import annotations
from abc import ABC, abstractmethod
from collections import defaultdict
import contextlib
import logging

from typing import Iterator, List

import more_itertools

from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.context.base import InferenceContextProvider
from metricflow.inference.context.data_warehouse import DataWarehouseInferenceContextProvider
from metricflow.inference.rule.base import InferenceRule
from metricflow.inference.solver.base import InferenceSolver
from metricflow.inference.renderer.base import InferenceRenderer


logger = logging.getLogger(__file__)


class InferenceProgressReporter(ABC):
    """Base class for reporting progress while running inference."""

    @staticmethod
    @abstractmethod
    @contextlib.contextmanager
    def warehouse() -> Iterator[None]:
        """Context manager that wraps the warehouse context fetching step."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    @contextlib.contextmanager
    def table(table: SqlTable, index: int, total: int) -> Iterator[None]:
        """Context manager that wraps context fetching for a single table."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    @contextlib.contextmanager
    def rules() -> Iterator[None]:
        """Context manager that wraps all rule-processing."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    @contextlib.contextmanager
    def solver() -> Iterator[None]:
        """Context manager that wraps the solving step."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    @contextlib.contextmanager
    def renderers() -> Iterator[None]:
        """Context manager that wraps the rendering step."""
        raise NotImplementedError


class NoOpInferenceProgressReporter(InferenceProgressReporter):
    """Pass-through implementation of `InferenceProgressReporter`."""

    @staticmethod
    @contextlib.contextmanager
    def warehouse() -> Iterator[None]:  # noqa: D
        yield

    @staticmethod
    @contextlib.contextmanager
    def table(table: SqlTable, index: int, total: int) -> Iterator[None]:  # noqa: D
        yield

    @staticmethod
    @contextlib.contextmanager
    def rules() -> Iterator[None]:  # noqa: D
        yield

    @staticmethod
    @contextlib.contextmanager
    def solver() -> Iterator[None]:  # noqa: D
        yield

    @staticmethod
    @contextlib.contextmanager
    def renderers() -> Iterator[None]:  # noqa: D
        yield


# TODO: we still need to add input/output context validations and optimizations.
# Case 1: Rule 1 requires Context of type A, but no provider privides it. Should fail before running
# Case 2: ContextProvider 1 provides Context of type A, but no rule uses it. Should proceed without actually fetching that context.
# Case 3: ContextProviders 1 and 2 both provide Contexts of type A, which is ambiguous. Should fail before running


class InferenceRunner:
    """Glues together all other inference classes in a sequence that actually runs inference."""

    def __init__(
        self,
        context_providers: List[InferenceContextProvider],
        ruleset: List[InferenceRule],
        solver: InferenceSolver,
        renderers: List[InferenceRenderer],
        progress_reporter: InferenceProgressReporter = NoOpInferenceProgressReporter(),
    ) -> None:
        """Initialize the inference runner.

        context_providers: a list of context providers to be used
        ruleset: the set of rules that will produce signals
        solver: the inference solver to be used
        renderers: the renderers that will write inference results as meaningful output
        progress_reporter: `InferenceProgressReporter` to report progress
        """

        logger.warning(
            "Data Source Inference is still in Beta. "
            "As such, you should not expect it to be 100% stable or be free of bugs. Any public CLI or Python interfaces may change without prior notice."
            " If you find any bugs or feel like something is not behaving as it should, feel free to open an issue on the Metricflow Github repo."
        )

        if len(context_providers) != 1 or not isinstance(context_providers[0], DataWarehouseInferenceContextProvider):
            raise ValueError("Currently, InferenceRunner requires exactly one DataWarehouseInferenceContextProvider.")

        if len(ruleset) == 0:
            raise ValueError("Running inference with an empty ruleset would produce no result.")

        if len(renderers) == 0:
            raise ValueError("Running inference with no renderer would discard the results.")

        self.context_providers = context_providers
        self.ruleset = ruleset
        self.solver = solver
        self.renderers = renderers
        self._progress = progress_reporter

    def run(self) -> None:
        """Runs inference with the given configs."""

        # FIXME: currently we only accept DataWarehouseContextProvider
        provider: DataWarehouseInferenceContextProvider = self.context_providers[0]  # type: ignore

        with self._progress.warehouse():
            warehouse = provider.get_context(
                table_progress=self._progress.table,
            )

        with self._progress.rules():
            signals_by_column = defaultdict(list)
            signals = [rule.process(warehouse) for rule in self.ruleset]
            for rule_signal in more_itertools.flatten(signals):
                signals_by_column[rule_signal.column].append(rule_signal)

        with self._progress.solver():
            results = [self.solver.solve_column(column, signals) for column, signals in signals_by_column.items()]

        with self._progress.renderers():
            for renderer in self.renderers:
                renderer.render(results)



from __future__ import annotations

from typing import Optional, List
from metricflow.model.objects.common import Metadata
from metricflow.model.objects.base import ModelWithMetadataParsing, HashableBaseModel
from metricflow.object_utils import ExtendedEnum, hash_strings
from metricflow.references import MeasureReference, TimeDimensionReference


class AggregationType(ExtendedEnum):
    """Aggregation methods for measures"""

    SUM = "sum"
    MIN = "min"
    MAX = "max"
    COUNT_DISTINCT = "count_distinct"
    # BOOLEAN is deprecated. Remove when customers have migrated.
    BOOLEAN = "boolean"
    SUM_BOOLEAN = "sum_boolean"
    AVERAGE = "average"

    # COUNT = "count" not yet implemented ... non-expansive as COUNT(COUNT(X), COUNT(Y)) != COUNT(COUNT(X ∪ Y))
    # AVERAGE = "average"  not yet implemented ...requires us to keep track of two quantities, count and sum

    @property
    def is_expansive(self) -> bool:
        """Expansive ≝ Op( X ∪ Y ∪ ...) = Op( Op(X) ∪ Op(Y) ∪ ...)"""
        return self in (AggregationType.SUM, AggregationType.MIN, AggregationType.MAX, AggregationType.BOOLEAN)

    @property
    def fill_nulls_with_0(self) -> bool:
        """Indicates if charts should show 0 instead of null where there are gaps in data."""
        return self in (AggregationType.SUM, AggregationType.COUNT_DISTINCT, AggregationType.SUM_BOOLEAN)

    @property
    def can_limit_dimension_values(self) -> bool:
        """Indicates if we can limit dimension values in charts.

        Currently, this means:
        1. The dimensions we care about most are the ones with the highest numeric values
        2. We can calculate the "other" column in the postprocessor (meaning the metric is expansive)
        """
        return self in (AggregationType.SUM, AggregationType.SUM_BOOLEAN)


class NonAdditiveDimensionParameters(HashableBaseModel):
    """Describes the params for specifying non-additive dimensions in a measure.

    NOTE: Currently, only TimeDimensions are supported for this filter
    """

    name: str
    window_choice: AggregationType
    window_groupings: List[str] = []

    @property
    def bucket_hash(self) -> str:
        """Returns the hash value used for grouping equivalent params."""
        values = [self.window_choice.name, self.name]
        values.extend(sorted(self.window_groupings))
        return hash_strings(values)


class Measure(HashableBaseModel, ModelWithMetadataParsing):
    """Describes a measure"""

    name: str
    agg: AggregationType
    description: Optional[str]
    create_metric: Optional[bool]
    expr: Optional[str] = None
    metadata: Optional[Metadata]
    non_additive_dimension: Optional[NonAdditiveDimensionParameters] = None

    # Defines the time dimension to aggregate this measure by. If not specified, it means to use the primary time
    # dimension in the data source.
    agg_time_dimension: Optional[str] = None

    @property
    def checked_agg_time_dimension(self) -> TimeDimensionReference:
        """Returns the aggregation time dimension, throwing an exception if it's not set."""
        assert (
            self.agg_time_dimension
        ), f"Aggregation time dimension for {self.name} should have been set during model transformation"
        return TimeDimensionReference(element_name=self.agg_time_dimension)

    @property
    def reference(self) -> MeasureReference:  # noqa: D
        return MeasureReference(element_name=self.name)



from __future__ import annotations

import logging
import time
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import Tuple, Sequence, Dict, List, Optional, FrozenSet

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import DimensionType, Dimension
from metricflow.model.objects.elements.identifier import IdentifierType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.object_utils import pformat_big_objects, flatten_nested_sequence
from metricflow.references import MeasureReference
from metricflow.specs import (
    LinklessIdentifierSpec,
    DEFAULT_TIME_GRANULARITY,
    LinkableSpecSet,
    DimensionSpec,
    TimeDimensionSpec,
    IdentifierSpec,
    MetricSpec,
)
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


class LinkableElementProperties(Enum):
    """The properties associated with a valid linkable element.

    Local means an element that is defined within the same data source as the measure. This definition is used
    throughout the related classes.
    """

    # A local element as per above definition.
    LOCAL = "local"
    # A local dimension that is prefixed with a local primary identifier.
    LOCAL_LINKED = "local_linked"
    # An element that was joined to the measure data source by an identifier.
    JOINED = "joined"
    # An element that was joined to the measure data source by joining multiple data sources.
    MULTI_HOP = "multi_hop"
    # A time dimension that is a version of a time dimension in a data source, but at a different granularity.
    DERIVED_TIME_GRANULARITY = "derived_time_granularity"
    # Refers to an identifier, not a dimension.
    IDENTIFIER = "identifier"
    # After an intersection operation.
    INTERSECTED = "intersected"

    @staticmethod
    def all_properties() -> FrozenSet[LinkableElementProperties]:  # noqa: D
        return frozenset(
            {
                LinkableElementProperties.LOCAL,
                LinkableElementProperties.LOCAL_LINKED,
                LinkableElementProperties.JOINED,
                LinkableElementProperties.MULTI_HOP,
                LinkableElementProperties.DERIVED_TIME_GRANULARITY,
            }
        )


@dataclass(frozen=True)
class JoinPathKey:
    """A key that can uniquely identify an element and the joins used to realize the element."""

    element_name: str
    identifier_links: Tuple[str, ...]
    time_granularity: Optional[TimeGranularity]


@dataclass(frozen=True)
class LinkableDimension:
    """Describes how a dimension can be realized by joining based on identifier links."""

    element_name: str
    identifier_links: Tuple[str, ...]
    properties: FrozenSet[LinkableElementProperties]
    time_granularity: Optional[TimeGranularity] = None

    @property
    def path_key(self) -> JoinPathKey:  # noqa: D
        return JoinPathKey(
            element_name=self.element_name,
            identifier_links=self.identifier_links,
            time_granularity=self.time_granularity,
        )

    @property
    def after_intersection(self) -> LinkableDimension:  # noqa: D
        return LinkableDimension(
            element_name=self.element_name,
            identifier_links=self.identifier_links,
            properties=frozenset({LinkableElementProperties.INTERSECTED}),
            time_granularity=self.time_granularity,
        )


@dataclass(frozen=True)
class LinkableIdentifier:
    """Describes how an identifier can be realized by joining based on identifier links."""

    element_name: str
    properties: FrozenSet[LinkableElementProperties]
    identifier_links: Tuple[str, ...]

    @property
    def path_key(self) -> JoinPathKey:  # noqa: D
        return JoinPathKey(
            element_name=self.element_name, identifier_links=self.identifier_links, time_granularity=None
        )

    @property
    def after_intersection(self) -> LinkableIdentifier:  # noqa: D
        return LinkableIdentifier(
            element_name=self.element_name,
            identifier_links=self.identifier_links,
            properties=frozenset({LinkableElementProperties.INTERSECTED}),
        )


@dataclass(frozen=True)
class LinkableElementSet:
    """Container class for storing all linkable elements for a metric."""

    linkable_dimensions: Tuple[LinkableDimension, ...]
    linkable_identifiers: Tuple[LinkableIdentifier, ...]

    # Ambiguous elements are ones where there are multiple join paths through different data sources that can be taken
    # to get the element. This currently represents an error when defining data sources.
    ambiguous_linkable_dimensions: Tuple[LinkableDimension, ...]
    ambiguous_linkable_identifiers: Tuple[LinkableIdentifier, ...]

    @staticmethod
    def merge(linkable_element_sets: Sequence[LinkableElementSet]) -> LinkableElementSet:
        """Combine multiple sets together.

        If there are elements with the same join key, those elements will be categorized as ambiguous.
        """
        key_to_linkable_dimension: Dict[JoinPathKey, List[LinkableDimension]] = defaultdict(list)
        key_to_linkable_identifier: Dict[JoinPathKey, List[LinkableIdentifier]] = defaultdict(list)

        for linkable_element_set in linkable_element_sets:
            for linkable_dimension in linkable_element_set.linkable_dimensions:
                key_to_linkable_dimension[linkable_dimension.path_key].append(linkable_dimension)
            for linkable_identifier in linkable_element_set.linkable_identifiers:
                key_to_linkable_identifier[linkable_identifier.path_key].append(linkable_identifier)

        linkable_dimensions = []
        linkable_identifiers = []

        ambiguous_linkable_dimensions = list(
            flatten_nested_sequence([x.ambiguous_linkable_dimensions for x in linkable_element_sets])
        )
        ambiguous_linkable_identifiers = list(
            flatten_nested_sequence([x.ambiguous_linkable_identifiers for x in linkable_element_sets])
        )

        for _, grouped_linkable_dimensions in key_to_linkable_dimension.items():
            for linkable_dimension in grouped_linkable_dimensions:
                if len(grouped_linkable_dimensions) == 1:
                    linkable_dimensions.append(linkable_dimension)
                else:
                    ambiguous_linkable_dimensions.append(linkable_dimension)

        for _, grouped_linkable_identifier in key_to_linkable_identifier.items():
            for linkable_identifier in grouped_linkable_identifier:
                if len(grouped_linkable_identifier) == 1:
                    linkable_identifiers.append(linkable_identifier)
                else:
                    ambiguous_linkable_identifiers.append(linkable_identifier)

        return LinkableElementSet(
            linkable_dimensions=tuple(linkable_dimensions),
            linkable_identifiers=tuple(linkable_identifiers),
            ambiguous_linkable_dimensions=tuple(ambiguous_linkable_dimensions),
            ambiguous_linkable_identifiers=tuple(ambiguous_linkable_identifiers),
        )

    @staticmethod
    def intersection(linkable_element_sets: Sequence[LinkableElementSet]) -> LinkableElementSet:
        """Find the intersection of all elements in the sets."""
        if len(linkable_element_sets) == 0:
            return LinkableElementSet(
                linkable_dimensions=(),
                linkable_identifiers=(),
                ambiguous_linkable_dimensions=(),
                ambiguous_linkable_identifiers=(),
            )

        # Need to find an easier syntax for finding the common items from multiple LinkableElementSets.
        common_linkable_dimensions = (
            set.intersection(
                *[set([y.after_intersection for y in x.linkable_dimensions]) for x in linkable_element_sets]
            )
            if len(linkable_element_sets) > 0
            else set()
        )

        common_linkable_identifiers = (
            set.intersection(
                *[set([y.after_intersection for y in x.linkable_identifiers]) for x in linkable_element_sets]
            )
            if len(linkable_element_sets) > 0
            else set()
        )

        common_ambiguous_linkable_dimensions = (
            set.intersection(
                *[set([y.after_intersection for y in x.ambiguous_linkable_dimensions]) for x in linkable_element_sets]
            )
            if len(linkable_element_sets) > 0
            else set()
        )

        common_ambiguous_linkable_identifiers = (
            set.intersection(
                *[set([y.after_intersection for y in x.ambiguous_linkable_identifiers]) for x in linkable_element_sets]
            )
            if len(linkable_element_sets) > 0
            else set()
        )

        return LinkableElementSet(
            linkable_dimensions=tuple(common_linkable_dimensions),
            linkable_identifiers=tuple(common_linkable_identifiers),
            ambiguous_linkable_dimensions=tuple(common_ambiguous_linkable_dimensions),
            ambiguous_linkable_identifiers=tuple(common_ambiguous_linkable_identifiers),
        )

    def filter(
        self, with_any_of: FrozenSet[LinkableElementProperties], without_any_of: FrozenSet[LinkableElementProperties]
    ) -> LinkableElementSet:
        """Filter elements in the set.

        First, only elements with at least one property in the "with_any_of" set are retained. Then, any elements with
        a property in "without_any_of" set are removed.
        """
        linkable_dimensions = tuple(
            x
            for x in self.linkable_dimensions
            if len(x.properties.intersection(with_any_of)) > 0 and len(x.properties.intersection(without_any_of)) == 0
        )

        linkable_identifiers = tuple(
            x
            for x in self.linkable_identifiers
            if len(x.properties.intersection(with_any_of)) > 0 and len(x.properties.intersection(without_any_of)) == 0
        )

        ambiguous_linkable_dimensions = tuple(
            x
            for x in self.ambiguous_linkable_dimensions
            if len(x.properties.intersection(with_any_of)) > 0 and len(x.properties.intersection(without_any_of)) == 0
        )

        ambiguous_linkable_identifiers = tuple(
            x
            for x in self.ambiguous_linkable_identifiers
            if len(x.properties.intersection(with_any_of)) > 0 and len(x.properties.intersection(without_any_of)) == 0
        )

        return LinkableElementSet(
            linkable_dimensions=linkable_dimensions,
            linkable_identifiers=linkable_identifiers,
            ambiguous_linkable_dimensions=ambiguous_linkable_dimensions,
            ambiguous_linkable_identifiers=ambiguous_linkable_identifiers,
        )

    @property
    def as_spec_set(self) -> LinkableSpecSet:  # noqa: D
        return LinkableSpecSet(
            dimension_specs=tuple(
                DimensionSpec(
                    element_name=x.element_name,
                    identifier_links=tuple(LinklessIdentifierSpec.from_element_name(x) for x in x.identifier_links),
                )
                for x in self.linkable_dimensions
                if not x.time_granularity
            ),
            time_dimension_specs=tuple(
                TimeDimensionSpec(
                    element_name=x.element_name,
                    identifier_links=tuple(LinklessIdentifierSpec.from_element_name(x) for x in x.identifier_links),
                    time_granularity=x.time_granularity,
                )
                for x in self.linkable_dimensions
                if x.time_granularity
            ),
            identifier_specs=tuple(
                IdentifierSpec(
                    element_name=x.element_name,
                    identifier_links=tuple(LinklessIdentifierSpec.from_element_name(x) for x in x.identifier_links),
                )
                for x in self.linkable_identifiers
            ),
        )


@dataclass(frozen=True)
class DataSourceJoinPathElement:
    """Describes joining a data source by the given identifier."""

    data_source: DataSource
    join_on_identifier: str


def _generate_linkable_time_dimensions(
    dimension: Dimension, identifier_links: Tuple[str, ...], with_properties: FrozenSet[LinkableElementProperties]
) -> Sequence[LinkableDimension]:
    """Generates different versions of the given dimension, but at other valid time granularities."""
    linkable_dimensions = []

    defined_time_granularity = (
        dimension.type_params.time_granularity if dimension.type_params else DEFAULT_TIME_GRANULARITY
    )
    for time_granularity in TimeGranularity:
        if time_granularity < defined_time_granularity:
            continue
        properties = set(with_properties)
        if time_granularity != defined_time_granularity:
            properties.add(LinkableElementProperties.DERIVED_TIME_GRANULARITY)

        linkable_dimensions.append(
            LinkableDimension(
                element_name=dimension.reference.element_name,
                identifier_links=identifier_links,
                time_granularity=time_granularity,
                properties=frozenset(properties),
            )
        )

    return linkable_dimensions


@dataclass(frozen=True)
class DataSourceJoinPath:
    """Describes a series of joins between the measure data source, and other data sources by identifier.

    For example:

    (measure_source JOIN dimension_source0 ON identifier A) JOIN dimension_source1 ON identifier B

    would be represented by 2 path elements [(data_source0, A), (dimension_source1, B)]
    """

    path_elements: Tuple[DataSourceJoinPathElement, ...]

    def create_linkable_element_set(self, with_properties: FrozenSet[LinkableElementProperties]) -> LinkableElementSet:
        """Given the current path, generate the respective linkable elements from the last data source in the path."""
        identifier_links = tuple(x.join_on_identifier for x in self.path_elements)

        assert len(self.path_elements) > 0
        data_source = self.path_elements[-1].data_source

        linkable_dimensions = []
        linkable_identifiers = []

        for dimension in data_source.dimensions:
            dimension_type = dimension.type
            if dimension_type == DimensionType.CATEGORICAL:
                linkable_dimensions.append(
                    LinkableDimension(
                        element_name=dimension.reference.element_name,
                        identifier_links=identifier_links,
                        properties=with_properties,
                    )
                )
            elif dimension_type == DimensionType.TIME:
                linkable_dimensions.extend(
                    _generate_linkable_time_dimensions(
                        dimension=dimension,
                        identifier_links=identifier_links,
                        with_properties=with_properties,
                    )
                )
            else:
                raise RuntimeError(f"Unhandled type: {dimension_type}")

        for identifier in data_source.identifiers:
            # Avoid creating "booking_id__booking_id"
            if identifier.reference.element_name != identifier_links[-1]:
                linkable_identifiers.append(
                    LinkableIdentifier(
                        element_name=identifier.reference.element_name,
                        identifier_links=identifier_links,
                        properties=with_properties.union({LinkableElementProperties.IDENTIFIER}),
                    )
                )

        return LinkableElementSet(
            linkable_dimensions=tuple(linkable_dimensions),
            linkable_identifiers=tuple(linkable_identifiers),
            ambiguous_linkable_dimensions=(),
            ambiguous_linkable_identifiers=(),
        )

    @property
    def last_data_source(self) -> DataSource:
        """The last data source that would be joined in this path."""
        assert len(self.path_elements) > 0
        return self.path_elements[-1].data_source


class ValidLinkableSpecResolver:
    """Figures out what linkable specs are valid for a given metric.

    e.g. Can you query the metric "bookings" by "listing__country_latest"?
    """

    def __init__(
        self,
        user_configured_model: UserConfiguredModel,
        max_identifier_links: int,
    ) -> None:
        """Constructor.

        Args:
            user_configured_model: the model to use.
            max_identifier_links: the maximum number of joins to do when computing valid elements.
        """
        self._user_configured_model = user_configured_model
        self._data_sources = self._user_configured_model.data_sources

        assert max_identifier_links >= 0
        self._max_identifier_links = max_identifier_links

        # Map measures / identifiers to data sources that contain them.
        self._identifier_to_data_source: Dict[str, List[DataSource]] = defaultdict(list)
        self._measure_to_data_source: Dict[str, List[DataSource]] = defaultdict(list)

        for data_source in self._data_sources:
            for identifier in data_source.identifiers:
                if identifier.type in (IdentifierType.PRIMARY, IdentifierType.UNIQUE):
                    self._identifier_to_data_source[identifier.reference.element_name].append(data_source)

        self._metric_to_linkable_element_sets: Dict[str, List[LinkableElementSet]] = {}

        start_time = time.time()
        for metric in self._user_configured_model.metrics:
            linkable_sets_for_measure = []
            for measure in metric.measure_references:
                linkable_sets_for_measure.append(self._get_linkable_element_set_for_measure(measure))

            self._metric_to_linkable_element_sets[metric.name] = linkable_sets_for_measure
        logger.info(f"Building the [metric -> valid linkable element] index took: {time.time() - start_time:.2f}s")

    def _get_data_source_for_measure(self, measure_reference: MeasureReference) -> DataSource:  # noqa: D
        data_sources_where_measure_was_found = []
        for data_source in self._data_sources:
            if any([x.reference.element_name == measure_reference.element_name for x in data_source.measures]):
                data_sources_where_measure_was_found.append(data_source)

        if len(data_sources_where_measure_was_found) == 0:
            raise ValueError(f"No data sources were found with {measure_reference} in the model")
        elif len(data_sources_where_measure_was_found) > 1:
            raise ValueError(
                f"Measure {measure_reference} was found in multiple data sources:\n"
                f"{pformat_big_objects(data_sources_where_measure_was_found)}"
            )
        return data_sources_where_measure_was_found[0]

    def _get_local_set(self, data_source: DataSource) -> LinkableElementSet:
        """Gets the local elements for a given data source."""
        linkable_dimensions = []
        linkable_identifiers = []

        for dimension in data_source.dimensions:
            dimension_type = dimension.type
            if dimension_type == DimensionType.CATEGORICAL:
                linkable_dimensions.append(
                    LinkableDimension(
                        element_name=dimension.reference.element_name,
                        identifier_links=(),
                        properties=frozenset({LinkableElementProperties.LOCAL}),
                    )
                )
            elif dimension_type == DimensionType.TIME:
                linkable_dimensions.extend(
                    _generate_linkable_time_dimensions(
                        dimension=dimension,
                        identifier_links=(),
                        with_properties=frozenset({LinkableElementProperties.LOCAL}),
                    )
                )
            else:
                raise RuntimeError(f"Unhandled type: {dimension_type}")

        additional_linkable_dimensions = []
        for identifier in data_source.identifiers:
            linkable_identifiers.append(
                LinkableIdentifier(
                    element_name=identifier.reference.element_name,
                    identifier_links=(),
                    properties=frozenset({LinkableElementProperties.LOCAL, LinkableElementProperties.IDENTIFIER}),
                )
            )
            # If a data source has a primary identifier, we allow users to query using the dundered syntax, even though
            # there is no join involved. e.g. in the test model, the "listings_latest" data source would allow using
            # "listing__country_latest" for the "listings" metric.
            if identifier.type == IdentifierType.PRIMARY:
                for linkable_dimension in linkable_dimensions:
                    properties = {LinkableElementProperties.LOCAL, LinkableElementProperties.LOCAL_LINKED}
                    additional_linkable_dimensions.append(
                        LinkableDimension(
                            element_name=linkable_dimension.element_name,
                            identifier_links=(identifier.reference.element_name,),
                            time_granularity=linkable_dimension.time_granularity,
                            properties=frozenset(linkable_dimension.properties.union(properties)),
                        )
                    )

        return LinkableElementSet(
            linkable_dimensions=tuple(linkable_dimensions + additional_linkable_dimensions),
            linkable_identifiers=tuple(linkable_identifiers),
            ambiguous_linkable_dimensions=(),
            ambiguous_linkable_identifiers=(),
        )

    def _get_data_sources_with_joinable_identifier(self, identifier_name: str) -> Sequence[DataSource]:
        # May switch to non-cached implementation.
        data_sources = self._identifier_to_data_source[identifier_name]
        return data_sources

    def _get_linkable_element_set_for_measure(self, measure_reference: MeasureReference) -> LinkableElementSet:
        """Get the valid linkable elements for the given measure."""
        measure_data_source = self._get_data_source_for_measure(measure_reference)

        # Create local elements
        local_linkable_elements = self._get_local_set(measure_data_source)
        join_paths = []

        # Create 1-hop elements
        for identifier in measure_data_source.identifiers:
            data_sources = self._get_data_sources_with_joinable_identifier(identifier.reference.element_name)
            for data_source in data_sources:
                if data_source.name == measure_data_source.name:
                    continue
                join_paths.append(
                    DataSourceJoinPath(
                        path_elements=(
                            DataSourceJoinPathElement(
                                data_source=data_source, join_on_identifier=identifier.reference.element_name
                            ),
                        )
                    )
                )
        all_linkable_elements = LinkableElementSet.merge(
            [local_linkable_elements]
            + [
                x.create_linkable_element_set(with_properties=frozenset({LinkableElementProperties.JOINED}))
                for x in join_paths
            ]
        )

        # Create multi-hop elements. At each iteration, we generate the list of valid elements based on the current join
        # path, extend all paths to include the next valid data source, then repeat.
        for i in range(self._max_identifier_links - 1):
            new_join_paths: List[DataSourceJoinPath] = []
            for join_path in join_paths:
                new_join_paths.extend(
                    self._find_next_possible_paths(measure_data_source=measure_data_source, current_join_path=join_path)
                )

            if len(new_join_paths) == 0:
                return all_linkable_elements

            all_linkable_elements = LinkableElementSet.merge(
                [all_linkable_elements]
                + [
                    x.create_linkable_element_set(
                        with_properties=frozenset(
                            {LinkableElementProperties.JOINED, LinkableElementProperties.MULTI_HOP}
                        )
                    )
                    for x in new_join_paths
                ]
            )
            join_paths = new_join_paths
        return all_linkable_elements

    def get_linkable_elements_for_metrics(
        self,
        metric_specs: Sequence[MetricSpec],
        with_any_of: FrozenSet[LinkableElementProperties],
        without_any_of: FrozenSet[LinkableElementProperties],
    ) -> LinkableElementSet:
        """Gets the valid linkable elements that are common to all requested metrics."""
        linkable_element_sets = []
        for metric_spec in metric_specs:
            element_sets = self._metric_to_linkable_element_sets.get(metric_spec.element_name)
            if not element_sets:
                raise ValueError(f"Unknown metric: {metric_spec} in element set")
            metric_result = LinkableElementSet.intersection(
                [x.filter(with_any_of=with_any_of, without_any_of=without_any_of) for x in element_sets]
            )
            linkable_element_sets.append(metric_result)

        return LinkableElementSet.intersection(linkable_element_sets)

    def _find_next_possible_paths(
        self, measure_data_source: DataSource, current_join_path: DataSourceJoinPath
    ) -> Sequence[DataSourceJoinPath]:
        """Generate the set of possible paths that are 1 data source join longer that the "current_join_path"."""
        data_source = current_join_path.last_data_source
        new_join_paths = []

        for identifier in data_source.identifiers:
            identifier_name = identifier.reference.element_name

            if identifier_name in set(x.join_on_identifier for x in current_join_path.path_elements):
                continue

            data_sources_that_can_be_joined = self._get_data_sources_with_joinable_identifier(identifier_name)
            for data_source in data_sources_that_can_be_joined:
                # Don't create cycles in the join path by not repeating a data source in the path.
                if data_source.name == measure_data_source.name or any(
                    tuple(x.data_source.name == data_source.name for x in current_join_path.path_elements)
                ):
                    continue

                new_join_path = DataSourceJoinPath(
                    path_elements=current_join_path.path_elements
                    + (DataSourceJoinPathElement(data_source=data_source, join_on_identifier=identifier_name),)
                )
                new_join_paths.append(new_join_path)

        return new_join_paths



import logging

from metricflow.column_assoc import ColumnAssociation, SingleColumnCorrelationKey
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import ReadSqlSourceNode, JoinToBaseOutputNode, JoinDescription
from metricflow.dataflow.sql_table import SqlTable
from metricflow.instances import (
    InstanceSet,
    MeasureInstance,
    DataSourceElementReference,
    AggregationState,
)
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.specs import (
    IdentifierSpec,
    MeasureSpec,
    TimeDimensionSpec,
    LinklessIdentifierSpec,
    DimensionSpec,
    InstanceSpecSet,
)
from metricflow.sql.sql_exprs import SqlColumnReferenceExpression, SqlColumnReference
from metricflow.sql.sql_plan import SqlSelectStatementNode, SqlSelectColumn, SqlTableFromClauseNode
from metricflow.time.time_granularity import TimeGranularity
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository

logger = logging.getLogger(__name__)


def test_no_parent_node_data_set(
    simple_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> None:
    """Tests getting the data set from a single node."""
    resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DefaultColumnAssociationResolver(simple_semantic_model),
        semantic_model=simple_semantic_model,
        time_spine_source=time_spine_source,
    )

    data_set = SqlDataSet(
        instance_set=InstanceSet(
            measure_instances=(
                MeasureInstance(
                    associated_columns=(
                        ColumnAssociation(column_name="bookings", column_correlation_key=SingleColumnCorrelationKey()),
                    ),
                    defined_from=(
                        DataSourceElementReference(
                            data_source_name="fct_bookings_data_source", element_name="bookings"
                        ),
                    ),
                    spec=MeasureSpec(
                        element_name="bookings",
                    ),
                    aggregation_state=AggregationState.NON_AGGREGATED,
                ),
            ),
            dimension_instances=(),
            time_dimension_instances=(),
            identifier_instances=(),
        ),
        sql_select_node=SqlSelectStatementNode(
            description="test0",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(SqlColumnReference(table_alias="src", column_name="bookings")),
                    column_alias="bookings",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="src",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
        ),
    )

    node = ReadSqlSourceNode(data_set=data_set)

    assert resolver.get_output_data_set(node).instance_set == data_set.instance_set


def test_joined_node_data_set(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> None:
    """Tests getting the data set from a dataflow plan with a join."""
    resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DefaultColumnAssociationResolver(simple_semantic_model),
        semantic_model=simple_semantic_model,
        time_spine_source=time_spine_source,
    )

    # Join "revenue" with "users_latest" to get "user__home_state_latest"
    revenue_node = consistent_id_object_repository.simple_model_read_nodes["revenue"]
    users_node = consistent_id_object_repository.simple_model_read_nodes["users_latest"]
    join_node = JoinToBaseOutputNode(
        parent_node=revenue_node,
        join_targets=[
            JoinDescription(
                join_node=users_node,
                join_on_identifier=LinklessIdentifierSpec.from_element_name("user"),
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            )
        ],
    )

    join_node_output_data_set = resolver.get_output_data_set(join_node)
    assert join_node_output_data_set.instance_set.spec_set == InstanceSpecSet(
        metric_specs=(),
        measure_specs=(
            MeasureSpec(
                element_name="txn_revenue",
            ),
        ),
        dimension_specs=(
            DimensionSpec(
                element_name="home_state_latest",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
            ),
        ),
        identifier_specs=(IdentifierSpec(element_name="user", identifier_links=()),),
        time_dimension_specs=(
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.DAY),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.WEEK),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.QUARTER),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.YEAR),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.DAY,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.WEEK,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.MONTH,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.QUARTER,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.YEAR,
            ),
        ),
    )



import logging

from _pytest.fixtures import FixtureRequest

from metricflow.instances import DataSourceReference
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs import (
    DimensionSpec,
    IdentifierSpec,
    InstanceSpecSet,
    MeasureSpec,
    TimeDimensionSpec,
    LinklessIdentifierSpec,
)
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.time.time_granularity import TimeGranularity
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.sql.compare_sql_plan import assert_rendered_sql_equal

logger = logging.getLogger(__name__)


def test_convert_table_data_source_without_measures(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    """Simple test for converting a table data source. Since there are no measures, primary time is not checked."""
    users_data_set = consistent_id_object_repository.simple_model_data_sets["users_latest"]

    expected_spec_set = InstanceSpecSet(
        metric_specs=(),
        measure_specs=(),
        dimension_specs=(
            DimensionSpec(element_name="home_state_latest", identifier_links=()),
            DimensionSpec(
                element_name="home_state_latest",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
            ),
        ),
        identifier_specs=(IdentifierSpec(element_name="user", identifier_links=()),),
        time_dimension_specs=(
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.DAY),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.WEEK),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.QUARTER),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.YEAR),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.DAY,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.WEEK,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.MONTH,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.QUARTER,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="user", identifier_links=()),),
                time_granularity=TimeGranularity.YEAR,
            ),
        ),
    )

    assert users_data_set.instance_set.spec_set == expected_spec_set
    assert users_data_set.data_source_reference == DataSourceReference(data_source_name="users_latest")
    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan_id="plan0",
        select_node=users_data_set.sql_select_node,
        sql_client=sql_client,
    )


def test_convert_table_data_source_with_measures(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    """Complete test of table data source conversion. This includes the full set of measures/identifiers/dimensions.

    Measures trigger a primary time dimension validation. Additionally, this includes both categorical and time
    dimension types, which should cover most, if not all, of the table source branches in the target class.
    """
    id_verifications_data_set = consistent_id_object_repository.simple_model_data_sets["id_verifications"]

    expected_spec_set = InstanceSpecSet(
        metric_specs=(),
        measure_specs=(MeasureSpec(element_name="identity_verifications"),),
        dimension_specs=(
            DimensionSpec(element_name="verification_type", identifier_links=()),
            DimensionSpec(
                element_name="verification_type",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
            ),
        ),
        identifier_specs=(
            IdentifierSpec(element_name="verification", identifier_links=()),
            IdentifierSpec(element_name="user", identifier_links=()),
            IdentifierSpec(
                element_name="user",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
            ),
        ),
        time_dimension_specs=(
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.DAY),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.WEEK),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.QUARTER),
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.YEAR),
            TimeDimensionSpec(element_name="ds_partitioned", identifier_links=(), time_granularity=TimeGranularity.DAY),
            TimeDimensionSpec(
                element_name="ds_partitioned", identifier_links=(), time_granularity=TimeGranularity.WEEK
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned", identifier_links=(), time_granularity=TimeGranularity.MONTH
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned", identifier_links=(), time_granularity=TimeGranularity.QUARTER
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned", identifier_links=(), time_granularity=TimeGranularity.YEAR
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.DAY,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.WEEK,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.MONTH,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.QUARTER,
            ),
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.YEAR,
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.DAY,
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.WEEK,
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.MONTH,
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.QUARTER,
            ),
            TimeDimensionSpec(
                element_name="ds_partitioned",
                identifier_links=(LinklessIdentifierSpec(element_name="verification", identifier_links=()),),
                time_granularity=TimeGranularity.YEAR,
            ),
        ),
    )

    assert id_verifications_data_set.instance_set.spec_set == expected_spec_set
    assert id_verifications_data_set.data_source_reference == DataSourceReference(data_source_name="id_verifications")
    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan_id="plan0",
        select_node=id_verifications_data_set.sql_select_node,
        sql_client=sql_client,
    )


def test_convert_query_data_source(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    default_sql_plan_renderer: SqlQueryPlanRenderer,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    bookings_data_set = consistent_id_object_repository.simple_model_data_sets["revenue"]

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan_id="plan0",
        select_node=bookings_data_set.sql_select_node,
        sql_client=sql_client,
    )



from typing import Dict, List, Set
from metricflow.instances import DataSourceElementReference

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.identifier import Identifier
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    ModelValidationRule,
    ValidationWarning,
    validate_safely,
    ValidationIssueType,
)
from metricflow.references import IdentifierReference


class CommonIdentifiersRule(ModelValidationRule):
    """Checks that identifiers exist on more than one data source"""

    @staticmethod
    def _map_data_source_identifiers(data_sources: List[DataSource]) -> Dict[IdentifierReference, Set[str]]:
        """Generate mapping of identifier names to the set of data_sources where it is defined"""
        identifiers_to_data_sources: Dict[IdentifierReference, Set[str]] = {}
        for data_source in data_sources or []:
            for identifier in data_source.identifiers or []:
                if identifier.reference in identifiers_to_data_sources:
                    identifiers_to_data_sources[identifier.reference].add(data_source.name)
                else:
                    identifiers_to_data_sources[identifier.reference] = {data_source.name}
        return identifiers_to_data_sources

    @staticmethod
    @validate_safely(whats_being_done="checking identifier exists on more than one data source")
    def _check_identifier(
        identifier: Identifier,
        data_source: DataSource,
        identifiers_to_data_sources: Dict[IdentifierReference, Set[str]],
    ) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []
        # If the identifier is the dict and if the set of data sources minus this data source is empty,
        # then we warn the user that their identifier will be unused in joins
        if (
            identifier.reference in identifiers_to_data_sources
            and len(identifiers_to_data_sources[identifier.reference].difference({data_source.name})) == 0
        ):
            issues.append(
                ValidationWarning(
                    context=DataSourceElementContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source_element=DataSourceElementReference(
                            data_source_name=data_source.name, element_name=identifier.name
                        ),
                        element_type=DataSourceElementType.IDENTIFIER,
                    ),
                    message=f"Identifier `{identifier.reference.element_name}` "
                    f"only found in one data source `{data_source.name}` "
                    f"which means it will be unused in joins.",
                )
            )
        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation warning if identifiers are only one one data source")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:
        """Issues a warning for any identifier that is associated with only one data_source"""
        issues = []

        identifiers_to_data_sources = CommonIdentifiersRule._map_data_source_identifiers(model.data_sources)
        for data_source in model.data_sources or []:
            for identifier in data_source.identifiers or []:
                issues += CommonIdentifiersRule._check_identifier(
                    identifier=identifier,
                    data_source=data_source,
                    identifiers_to_data_sources=identifiers_to_data_sources,
                )

        return issues



from metricflow.dataflow.sql_column import SqlColumn
from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.context.data_warehouse import (
    DataWarehouseInferenceContext,
    ColumnProperties,
    InferenceColumnType,
    TableProperties,
)
from metricflow.inference.models import InferenceSignalConfidence, InferenceSignalType
from metricflow.inference.rule.rules import ColumnMatcherRule, LowCardinalityRatioRule


def create_context_with_counts(rows: int, distinct: int, nulls: int) -> DataWarehouseInferenceContext:
    """Get a `DataWarehouseInferenceContext` with the designated counts"""
    return DataWarehouseInferenceContext(
        table_props=[
            TableProperties(
                table=SqlTable.from_string("db.schema.table"),
                column_props=[
                    ColumnProperties(
                        column=SqlColumn.from_string("db.schema.table.column"),
                        type=InferenceColumnType.INTEGER,
                        row_count=rows,
                        distinct_row_count=distinct,
                        null_count=nulls,
                        is_nullable=nulls != 0,
                        min_value=0,
                        max_value=rows - 1,
                    )
                ],
            )
        ]
    )


class TestLowCardinalityRule(LowCardinalityRatioRule):  # noqa: D
    type_node = InferenceSignalType.DIMENSION.CATEGORICAL
    confidence = InferenceSignalConfidence.MEDIUM
    only_applies_to_parent_signal = False


def test_column_matcher(warehouse_ctx: DataWarehouseInferenceContext) -> None:  # noqa: D
    class TestRule(ColumnMatcherRule):
        type_node = InferenceSignalType.DIMENSION.UNKNOWN
        confidence = InferenceSignalConfidence.MEDIUM
        match_reason = "test reason"
        only_applies_to_parent_signal = False

        def match_column(self, props: ColumnProperties) -> bool:
            return props.column.column_name.endswith("test_column")

    signals = TestRule().process(warehouse_ctx)

    assert len(signals) == 1
    assert signals[0].confidence == InferenceSignalConfidence.MEDIUM
    assert signals[0].type_node == InferenceSignalType.DIMENSION.UNKNOWN
    assert signals[0].column == SqlColumn.from_string("db.schema.table.test_column")
    assert signals[0].reason == "test reason"
    assert not signals[0].only_applies_to_parent


def test_low_cardinality_ratio_rule_high_cardinality_doesnt_match() -> None:  # noqa: D
    rule = TestLowCardinalityRule(0.1)
    ctx = create_context_with_counts(100, 100, 0)

    signals = rule.process(ctx)
    assert len(signals) == 0


def test_low_cardinality_ratio_rule_low_cardinality_lots_of_nulls_doesnt_match() -> None:  # noqa: D
    rule = TestLowCardinalityRule(0.1)
    ctx = create_context_with_counts(100, 2, 99)

    signals = rule.process(ctx)
    assert len(signals) == 0


def test_low_cardinality_ratio_rule_low_cardinality_all_nulls_doesnt_match() -> None:  # noqa: D
    rule = TestLowCardinalityRule(0.1)
    ctx = create_context_with_counts(100, 1, 100)

    signals = rule.process(ctx)
    assert len(signals) == 0


def test_low_cardinality_ratio_rule_low_cardinality_matches() -> None:  # noqa: D
    rule = TestLowCardinalityRule(0.1)
    ctx = create_context_with_counts(100, 1, 0)

    signals = rule.process(ctx)
    assert len(signals) == 1



from typing import List

from metricflow.inference.context.data_warehouse import (
    ColumnProperties,
    InferenceColumnType,
    DataWarehouseInferenceContext,
)
from metricflow.inference.models import InferenceSignalConfidence, InferenceSignalType, InferenceSignal
from metricflow.inference.rule.base import InferenceRule
from metricflow.inference.rule.rules import ColumnMatcherRule, LowCardinalityRatioRule


# -------------
# IDENTIFIERS
# -------------


class AnyIdentifierByNameRule(ColumnMatcherRule):
    """Inference rule that checks for columns ending with `id`.

    We searched for words ending with "id" just to assess the chance of this resulting in a
    false positive. Our guess is most of those words would rarely, if ever, be used as column names.
    Therefore, not adding a mandatory "_" before "id" would benefit the product by matching names
    like "userid", despite the rare "squid", "mermaid" or "android" matches.

    See: https://www.thefreedictionary.com/words-that-end-in-id

    It will always produce ID.UNKNOWN signal with VERY_HIGH confidence.
    """

    type_node = InferenceSignalType.ID.UNKNOWN
    confidence = InferenceSignalConfidence.VERY_HIGH
    only_applies_to_parent_signal = False
    match_reason = "Column name ends with `id`"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.column.column_name.lower().endswith("id")


class PrimaryIdentifierByNameRule(ColumnMatcherRule):
    """Inference rule that matches primary identifiers by their names.

    It will match columns such as `db.schema.mytable.mytable_id`,
    `db.schema.mytable.mytableid` and `db.schema.mytable.id`.

    It will always produce a ID.PRIMARY signal with VERY_HIGH confidence.
    """

    type_node = InferenceSignalType.ID.PRIMARY
    confidence = InferenceSignalConfidence.VERY_HIGH
    only_applies_to_parent_signal = False
    match_reason = "Column name matches `(table_name?)(_?)id`"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        col_lower = props.column.column_name.lower()
        table_lower = props.column.table_name.lower().rstrip("s")

        if col_lower == "id":
            return True

        return col_lower == f"{table_lower}_id" or col_lower == f"{table_lower}id"


class UniqueIdentifierByDistinctCountRule(ColumnMatcherRule):
    """Inference rule that matches unique identifiers by their COUNT DISTINCT.

    It will always produce a ID.UNIQUE complementary signal with VERY_HIGH confidence.
    """

    type_node = InferenceSignalType.ID.UNIQUE
    confidence = InferenceSignalConfidence.VERY_HIGH
    only_applies_to_parent_signal = True
    match_reason = "The values in the column are unique"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.distinct_row_count == props.row_count


class ForeignIdentifierByCardinalityRatioRule(LowCardinalityRatioRule):
    """Inference rule that checks for low cardinality columns.

    It will always produce ID.FOREIGN with MEDIUM confidence (complementary).
    """

    type_node = InferenceSignalType.ID.FOREIGN
    confidence = InferenceSignalConfidence.MEDIUM
    only_applies_to_parent_signal = True


# -------------
# DIMENSIONS
# -------------


class TimeDimensionByTimeTypeRule(ColumnMatcherRule):
    """Inference rule that checks for time (time, date, datetime, timestamp) columns.

    It will always produce DIMENSION.TIME with VERY_HIGH confidence.
    """

    type_node = InferenceSignalType.DIMENSION.TIME
    confidence = InferenceSignalConfidence.VERY_HIGH
    only_applies_to_parent_signal = False
    match_reason = "Column type is time (TIME, DATE, DATETIME, TIMESTAMP)"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.type == InferenceColumnType.DATETIME


class PrimaryTimeDimensionByNameRule(ColumnMatcherRule):
    """Inference rule that checks if the column name is one of `ds` or `created_at`

    It will always produce DIMENSION.PRIMARY_TIME with VERY_HIGH confidence.
    """

    type_node = InferenceSignalType.DIMENSION.PRIMARY_TIME
    confidence = InferenceSignalConfidence.VERY_HIGH
    only_applies_to_parent_signal = False
    match_reason = "Column name is either of 'ds', 'created_at', 'created_date' or 'created_time'"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.column.column_name in ["ds", "created_at", "created_date", "created_time"]


class PrimaryTimeDimensionIfOnlyTimeRule(InferenceRule):
    """Inference rule for checking if the column is the only time column in the table.

    It will always produce DIMENSION.PRIMARY_TIME signal with VERY_HIGH confidence.
    """

    def process(self, warehouse: DataWarehouseInferenceContext) -> List[InferenceSignal]:  # noqa: D
        signals: List[InferenceSignal] = []
        for table_props in warehouse.tables.values():
            time_cols = [
                col for col, col_props in table_props.columns.items() if col_props.type == InferenceColumnType.DATETIME
            ]
            if len(time_cols) == 1:
                signals.append(
                    InferenceSignal(
                        column=time_cols[0],
                        type_node=InferenceSignalType.DIMENSION.PRIMARY_TIME,
                        only_applies_to_parent=False,
                        reason="The column is the only time column in its table",
                        confidence=InferenceSignalConfidence.VERY_HIGH,
                    )
                )
        return signals


class CategoricalDimensionByBooleanTypeRule(ColumnMatcherRule):
    """Inference rule that checks for boolean columns.

    It will always produce DIMENSION.CATEGORICAL with VERY_HIGH confidence.
    """

    type_node = InferenceSignalType.DIMENSION.CATEGORICAL
    confidence = InferenceSignalConfidence.VERY_HIGH
    only_applies_to_parent_signal = False
    match_reason = "Column type is BOOLEAN"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.type == InferenceColumnType.BOOLEAN


class CategoricalDimensionByStringTypeRule(ColumnMatcherRule):
    """Inference rule that checks for string columns.

    It will always produce DIMENSION.CATEGORICAL with MEDIUM confidence (complementary).
    """

    type_node = InferenceSignalType.DIMENSION.CATEGORICAL
    confidence = InferenceSignalConfidence.MEDIUM
    only_applies_to_parent_signal = True
    match_reason = "Column type is STRING"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.type == InferenceColumnType.STRING


class CategoricalDimensionByIntegerTypeRule(ColumnMatcherRule):
    """Inference rule that checks for integer columns.

    It will always produce DIMENSION.CATEGORICAL with MEDIUM confidence (complementary).
    """

    type_node = InferenceSignalType.DIMENSION.CATEGORICAL
    confidence = InferenceSignalConfidence.MEDIUM
    only_applies_to_parent_signal = True
    match_reason = "Column type is INTEGER"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.type == InferenceColumnType.INTEGER


class CategoricalDimensionByCardinalityRatioRule(LowCardinalityRatioRule):
    """Inference rule that checks for low cardinality columns.

    It will always produce DIMENSION.CATEGORICAL with MEDIUM confidence (complementary).
    """

    type_node = InferenceSignalType.DIMENSION.CATEGORICAL
    confidence = InferenceSignalConfidence.MEDIUM
    only_applies_to_parent_signal = True


# -------------
# MEASURES
# -------------


class MeasureByRealTypeRule(ColumnMatcherRule):
    """Inference rule that checks for real (float, double) columns.

    It will always produce MEASURE with VERY_HIGH confidence.
    """

    type_node = InferenceSignalType.MEASURE.UNKNOWN
    confidence = InferenceSignalConfidence.VERY_HIGH
    only_applies_to_parent_signal = False
    match_reason = "Column type is real (FLOAT, DOUBLE, DOUBLE PRECISION)"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.type == InferenceColumnType.FLOAT


class MeasureByIntegerTypeRule(ColumnMatcherRule):
    """Inference rule that checks for integer  columns.

    It will always produce MEASURE with MEDIUM confidence (complementary).
    """

    type_node = InferenceSignalType.MEASURE.UNKNOWN
    confidence = InferenceSignalConfidence.MEDIUM
    only_applies_to_parent_signal = True
    match_reason = "Column type is INTEGER"

    def match_column(self, props: ColumnProperties) -> bool:  # noqa: D
        return props.type == InferenceColumnType.INTEGER


DEFAULT_RULESET = [
    AnyIdentifierByNameRule(),
    PrimaryIdentifierByNameRule(),
    UniqueIdentifierByDistinctCountRule(),
    ForeignIdentifierByCardinalityRatioRule(0.4),
    TimeDimensionByTimeTypeRule(),
    PrimaryTimeDimensionByNameRule(),
    PrimaryTimeDimensionIfOnlyTimeRule(),
    CategoricalDimensionByBooleanTypeRule(),
    CategoricalDimensionByStringTypeRule(),
    CategoricalDimensionByIntegerTypeRule(),
    CategoricalDimensionByCardinalityRatioRule(0.2),
    MeasureByRealTypeRule(),
    MeasureByIntegerTypeRule(),
]



from collections import defaultdict
from typing import Callable, Dict, List, Optional

from metricflow.dataflow.sql_column import SqlColumn
from metricflow.inference.models import (
    InferenceResult,
    InferenceSignal,
    InferenceSignalConfidence,
    InferenceSignalNode,
    InferenceSignalType,
)
from metricflow.inference.solver.base import InferenceSolver


NodeWeighterFunction = Callable[[InferenceSignalConfidence], int]


class WeightedTypeTreeInferenceSolver(InferenceSolver):
    """Assigns weights to each type in the column type tree and attempts to traverse it using a weight percentage threshold."""

    @staticmethod
    def default_weighter_function(confidence: InferenceSignalConfidence) -> int:
        """The default weighter function.

        It assigns weights 1, 2, 4 and 8 for LOW, MEDIUM, HIGH and VERY_HIGH confidences, respectively. It then sums
        the weights of all provided confidence scores.
        """
        confidence_weight_map = {
            InferenceSignalConfidence.LOW: 1,
            InferenceSignalConfidence.MEDIUM: 2,
            InferenceSignalConfidence.HIGH: 4,
            InferenceSignalConfidence.VERY_HIGH: 8,
        }

        return confidence_weight_map[confidence]

    def __init__(
        self, weight_percent_threshold: float = 0.6, weighter_function: Optional[NodeWeighterFunction] = None
    ) -> None:
        """Initialize the solver.

        weight_percent_threshold: a number in (0.5, 1]. If a node's weight corresponds to a percentage
            above this threshold with respect to its siblings' total weight sum, the solver will progress deeper
            into the type tree, entering that node. If not, it stops at the parent.
        weighter_function: a function that returns a weight given a confidence score. It will be used
            to assign integer weights to each node in the type tree based on its input signals.
        """
        assert (
            weight_percent_threshold > 0.5 and weight_percent_threshold <= 1
        ), f"weight_percent_threshold is {weight_percent_threshold}, but it must be > 0.5 and <= 1!"
        self._weight_percent_threshold = weight_percent_threshold

        self._weighter_function = (
            weighter_function
            if weighter_function is not None
            else WeightedTypeTreeInferenceSolver.default_weighter_function
        )

    def _get_cumulative_weights_for_root(
        self,
        root: InferenceSignalNode,
        output_weights: Dict[InferenceSignalNode, int],
        output_parent_only_weights: Dict[InferenceSignalNode, int],
        signals_by_type: Dict[InferenceSignalNode, List[InferenceSignal]],
    ) -> Dict[InferenceSignalNode, int]:
        """Get a dict of cumulative weights, starting at `root`.

        A parent node's weight is the sum of all its children plus its own weight. Children tagged as only
        applying to their parent are excluded from their grand-parent's sum.

        root: the root to start assigning cumulative weights from.
        output_weights: the output dictionary to assign the cumulative weights to
        output_parent_only_weights: similar to output_weights, but the weight of each node excludes the weights
            of all of its parent-only children
        signals_by_type: a dictionary that maps signal type nodes to signals
        """
        for child in root.children:
            self._get_cumulative_weights_for_root(
                root=child,
                output_weights=output_weights,
                output_parent_only_weights=output_parent_only_weights,
                signals_by_type=signals_by_type,
            )

        output_weights[root] = sum(self._weighter_function(signal.confidence) for signal in signals_by_type[root])
        output_weights[root] += sum(output_parent_only_weights[child] for child in root.children)

        output_parent_only_weights[root] = sum(
            self._weighter_function(signal.confidence)
            for signal in signals_by_type[root]
            if not signal.only_applies_to_parent
        )

        return output_weights

    def _get_cumulative_weights(self, signals: List[InferenceSignal]) -> Dict[InferenceSignalNode, int]:
        """Get the cumulative weights dict for a list of signals"""
        signals_by_type: Dict[InferenceSignalNode, List[InferenceSignal]] = defaultdict(list)
        for signal in signals:
            signals_by_type[signal.type_node].append(signal)

        return self._get_cumulative_weights_for_root(
            root=InferenceSignalType.UNKNOWN,
            output_weights=defaultdict(int),
            output_parent_only_weights=defaultdict(int),
            signals_by_type=signals_by_type,
        )

    def solve_column(self, column: SqlColumn, signals: List[InferenceSignal]) -> InferenceResult:
        """Find the appropriate type for a column by traversing through the type tree.

        It traverses the tree by giving weights to all nodes and greedily finding the path with the most
        weight until it either finds a leaf or there is a "weight bifurcation" in the path with respect
        to the provided `weight_percent_threshold`.
        """
        if len(signals) == 0:
            return InferenceResult(
                column=column,
                type_node=InferenceSignalType.UNKNOWN,
                reasons=[],
                problems=[
                    "No signals were extracted for this column",
                    "Inference solver could not determine a type for this column",
                ],
            )

        reasons_by_type: Dict[InferenceSignalNode, List[str]] = defaultdict(list)
        for signal in signals:
            reasons_by_type[signal.type_node].append(f"{signal.reason} ({signal.type_node.name})")

        node_weights = self._get_cumulative_weights(signals)

        reasons: List[str] = []
        problems: List[str] = []
        node = InferenceSignalType.UNKNOWN
        while len(node.children) > 0:
            children_weight_total = sum(node_weights[child] for child in node.children)

            if children_weight_total == 0:
                break

            next_node = None
            for child in node.children:
                if node_weights[child] / children_weight_total >= self._weight_percent_threshold:
                    next_node = child
                    reasons += reasons_by_type[child]
                    break

            if next_node is None:
                if len(node.children) > 0:  # there was confusion
                    children_weight_strings = [
                        f"{child.name} (weight {node_weights[child]})"
                        for child in node.children
                        if node_weights[child] != 0
                    ]
                    child_str = " / ".join(children_weight_strings)
                    problems.append(f"Solver is confused between {child_str}")
                break

            node = next_node

        if node == InferenceSignalType.UNKNOWN:
            problems.append("Inference solver could not determine a type for this column")

        return InferenceResult(column=column, type_node=node, reasons=reasons, problems=problems)



from typing import Dict, List
from metricflow.instances import DataSourceReference

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    FileContext,
    ModelValidationRule,
    DataSourceElementType,
    ValidationError,
    ValidationIssueType,
    validate_safely,
)


class ElementConsistencyRule(ModelValidationRule):
    """Checks that elements in data sources with the same name are of the same element type across the model"""

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring model wide element consistency")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues = []
        for data_source in model.data_sources:
            issues += ElementConsistencyRule._check_data_source(model=model, data_source=data_source, add_to_dict=True)
        return issues

    @staticmethod
    def _check_element_type(
        name_to_type: Dict[str, DataSourceElementType],
        data_source: DataSource,
        element_name: str,
        element_type: DataSourceElementType,
        add_to_dict: bool,
    ) -> List[ValidationIssueType]:
        """Check if the given element matches the expected type.

        Args:
            name_to_type: dict from the name of the element to the expected type of the element in the model.
            data_source: the data source on which the element exists
            element_name: name of the element
            element_type: the type of the element
            add_to_dict: if the given element does not exist in the dictionary, whether to add it.
        """
        issues: List[ValidationIssueType] = []
        if element_name in name_to_type:
            existing_type = name_to_type[element_name]
            if existing_type != element_type:
                issues.append(
                    ValidationError(
                        context=DataSourceContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source=DataSourceReference(data_source_name=data_source.name),
                        ),
                        message=f"In data source {data_source.name}, element `{element_name}` is of type "
                        f"{element_type}, but it was previously used earlier in the model as "
                        f"{name_to_type[element_name]}",
                    )
                )
        else:
            if add_to_dict:
                name_to_type[element_name] = element_type
            elif element_type != DataSourceElementType.DIMENSION:
                # TODO: Can't check dimensions effectively as their name changes.
                issues.append(
                    ValidationError(
                        context=DataSourceContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source=DataSourceReference(data_source_name=data_source.name),
                        ),
                        message=f"In data source {data_source.name}, the element named {element_name} "
                        f"of type {element_type} is not known in the model.",
                    )
                )

        return issues

    @staticmethod
    def _get_element_types(model: UserConfiguredModel) -> Dict[str, DataSourceElementType]:
        # Store the element types
        element_types: Dict[str, DataSourceElementType] = {}
        for data_source in model.data_sources:
            if data_source.measures:
                for measure in data_source.measures:
                    element_types[measure.name] = DataSourceElementType.MEASURE
            if data_source.dimensions:
                for dimension in data_source.dimensions:
                    element_types[dimension.name] = DataSourceElementType.DIMENSION
            if data_source.identifiers:
                for identifier in data_source.identifiers:
                    element_types[identifier.name] = DataSourceElementType.IDENTIFIER
        return element_types

    @staticmethod
    @validate_safely(
        whats_being_done="checking that a data source's elements that share names with other elements across the model also are the same type"
    )
    def _check_data_source(
        model: UserConfiguredModel, data_source: DataSource, add_to_dict: bool
    ) -> List[ValidationIssueType]:  # noqa: D
        """Check if the elements in the data source matches the expected type.

        Args:
            model: UserConfiguredModel to check
            data_source: the data source to check
            add_to_dict: if the given element does not exist in the dictionary, whether to add it.

        Returns:
            A list of validation issues found with elements of data sources for the model
        """
        measure_name_tuples = [(x.name, DataSourceElementType.MEASURE) for x in data_source.measures or []]
        dimension_name_tuples = [(x.name, DataSourceElementType.DIMENSION) for x in data_source.dimensions or []]
        identifier_name_tuples = [(x.name, DataSourceElementType.IDENTIFIER) for x in data_source.identifiers or []]
        issues = []
        for element_name, element_type in measure_name_tuples + dimension_name_tuples + identifier_name_tuples:
            issues += ElementConsistencyRule._check_element_type(
                name_to_type=ElementConsistencyRule._get_element_types(model),
                data_source=data_source,
                element_name=element_name,
                element_type=element_type,
                add_to_dict=add_to_dict,
            )
        return issues
